pro LM_sky_list, LM_sky
;
;   LM night sky feature list - returns structure
;
;   order is wavelength, then 0/1 for each spectroscopic mode
;     0=don't use, 1 = use
;
;   spec modes are : JHK, K, J-2nd, J-3rd, LM, HK, BVR/VPH
;
;   DPC  20090221
;
;   most wavelengths from Steve Lord's ATRAN program
;
nLM_modes = 7
;
LM_line_and_quality = [ $
           ;
           ; optical lines for PRISM first
           ;
              ;
           ;
           ;    Z-Band lines
           ;
              ;
          ;
          ;  J-Band lines
          ;
              ;
          ;
          ;   J-H Atmosphere abs
          ;
              ;
          ;
          ;  H-Band
          ;
              ;
          ;
          ;  H-K Atmosphere abs
          ;
              ;
          ;
          ;  K-s Band
          ;
              ;
          ;
          ; K-l Band
          ;
              ;2.3851540d0, 1, 1, 0, 0, 0, 1,0, $
              ;2.3973060d0, 1, 1, 0, 0, 0, 1,0]
          ;
          ;	LM bands
          ;
          	3.315940d0, 0, 0, 0, 0, 1, 0, 0, $
          	3.678280d0, 0, 0, 0, 0, 1, 0, 0, $
          	3.885000d0, 0, 0, 0, 0, 0, 0, 0, $
          	-3.90590d0, 0, 0, 0, 0, 1, 0, 0, $
          	3.923760d0, 0, 0, 0, 0, 0, 0, 0, $
          	-4.499360d0, 0, 0, 0, 0, 1, 0, 0, $	; minus for absorption feature
          	4.814210d0, 0, 0, 0, 0, 1, 0, 0, $
          	4.896040d0, 0, 0, 0, 0, 1, 0, 0, $
          	4.959250d0, 0, 0, 0, 0, 1, 0, 0, $
          	5.021217d0, 0, 0, 0, 0, 0, 0, 0,  $
          	5.145180d0, 0, 0, 0, 0, 0, 0, 0,  $
          	5.208300d0, 0, 0, 0, 0, 0, 0, 0,  $
          	5.352950d0, 0, 0, 0, 0, 0, 0, 0,  $
          	5.420800d0, 0, 0, 0, 0, 0, 0, 0,  $
          	5.472350d0, 0, 0, 0, 0, 1, 0, 0  ]
          	;
;
;   parse into line and quality arrays
;
LM_size = N_ELEMENTS(LM_line_and_quality)
nlines = LM_size / (nLM_modes + 1)
LM_line = dblarr(nlines)
LM_quality = intarr(nLM_modes, nlines)    ; (spec_mode, line)
;
for iline = 0, nlines-1 do begin
    jline = iline*(nLM_modes + 1)
    LM_line[iline] = LM_line_and_quality[jline]
    for spec_mode = 0, nLM_modes-1 do begin
       LM_quality[spec_mode, iline] = LM_line_and_quality[jline+spec_mode+1]
    endfor
endfor
;
nLM_lines = N_ELEMENTS(LM_line)
;
LM_sky = {nlines: nLM_lines, wave: LM_line, qual: LM_quality}
;
end