;
; IDL Event Callback Procedures
; MSP_WCT_eventcb
;
; v2.3.0 20111201 - added FITS header variables to reflect slit, et al.
; v2.2.2 20111130 - removed STATUS=status in call to GAUSSFIT - hmmm
; v2.2.1 20111129 - fixed file naming errors, added full path qualification to summary.txt call
; v2.2.0 20110804 - new Argon line list
; v2.1.4 20100917 - more modes supported
; v2.0.1 20100714 - minor bug fixes
; v2.0.0 20100115 - stripped out GPU stuff, added version string
; v2.1.0 20100910 - added K17-Spec wavelengths and full support
; v2.1.1 20100917 - corrected J-Spec Order designations (up by one order each)
; v2.1.2 20100917 - added IZYJ-Spec modes
;
;
; Empty stub procedure used for autoloading.
;
pro MSP_WCT_eventcb

end

;------------------------------------- Basic Set Up ---------
pro WCT_startup, wWidget, _EXTRA=_VWBExtra_                ; called on PostCreation
    ;print, "Changing font to Helvetica", only if windows
    if(STRTRIM(STRUPCASE(!VERSION.OS_FAMILY),2) eq 'WINDOWS') then begin
    	WIDGET_CONTROL, DEFAULT_FONT='Tahoma*13'
    endif
    ;
    version_short_string = 'v2.3.0 20111201 DPC'
	version_string = 'MSP_WCT (Mimir Software Programs - Wavelength Calibration Tool) '+version_short_string
	;
    ;
    ;   get line lists as structures
    ;
    Argon_line_list, Argon
    OH_line_list, OH
    LM_sky_list, LM_sky
    ;
    jhk_jhk_dispersion =    [1.0180d0,    1.3786d-3,      8.59d-8]          ;  JHK-Spec
    jhk_hk_dispersion =     [1.0180d0,    1.3786d-3,      8.59d-8]			;  HK-Spec
    jhk_k_dispersion =      [0.99373d0,   1.41551d-3,  6.51327d-8]          ;  1st order K-spec
    lm_1st_dispersion  =    [2.3395d0,    3.29d-3,   3.80e-8]               ;  LM-Spec
    lm_3rd_dispersion =     [0.790464d0,  1.20872d-3, -3.29515d-8]		    ;  J-Spec 3rd
    lm_4th_dispersion =     [0.61114d0,  8.40087d-4, 5.18533d-8]			;  J-Spec 4th
    sed_dispersion =        [0.0d0,         0.0d0,          0.0d0]          ;  tbd
    bvr_vph_dispersion =    [0.0d0,      0.0d0,       0.0d0]                ; tbd
    lm_2nd_k17_dispersion = [1.7994018d0, 5.29451d-4, 1.10668d-8]			; K17-Spec
    jhk_izyj_1st_dispersion = [1.01555d0,    1.37564d-3,      1.0911d-7]	; IZYJ-Spec 1st
    jhk_izyj_2nd_dispersion = [0.515484d0,   6.94212d-4,      4.55861d-8]   ; IZYJ-Spec 2nd
    jhk_h17_dispersion =    [1.515922d0,    4.45832e-4, 6.25315e-9]			; H17-Spec
    jhk_iz17_2nd_dispersion = [0.8001667d0, 1.3519d-4, 6.103684d-8]			; IZ17
    ;
    ;
    ;   create and clear image and coadd structure
    ;
    WCT_struct = {version_short_string: version_short_string, $
    			  version_string: version_string, $
    			  instrument: 0, $                  ; 0=Mimir, 1=PRISM
                  camera: 0,$                       ; 0=F/5, 1=F/17
                  slit: 1, $                        ; Mimir slits: 0, 1, 4, 11; Prism slits: 0, 1
                  object: 0, $                      ; 0=single, >0 = multi-object
                  spec_mode: -1, $                  ; 0 = JHK; 1 = K-Spec; 2 = J-Spec 3rd Order;
                                                    ;   3=J-Spec 4th Order, 4 = LM Spec, 5=HK Spec, 6 = K17-spec,
                                                    ;   7 = IZYJ-1st , 8 = IZYJ-2nd, 9 = BVR/VPH, -1=none set yet
                  debug_flag: 0, $                  ; 1 = turn on debugging for indicated step, 0 = turn off debug
                  current_step: 0, $                ; saves current processing step number
                  wave_limits: dblarr(15,2), $      ; array of start/stop wavelength fit limits, as (spec_mode, start/stop)
                  cal_mode:  -1, $                  ; 0 = Argon; 1 = OH night sky lines; 2 = LM night sky; -1 = none set
                  deleted_value: -100.0, $       	; pixel value for deleted data
                  ;
                  file_mode: -1, $                  ; 0 = single image, non-zero = multiple images
                  file_name: '', $                  ; filename if single
                  files_multi: strarr(250), $       ; filenames buffer for multiple
                  nfiles: 0, $                      ; number of file names in files_multi buffer
                  coadd:     0, $                   ; 0 = done or not needed; else = need to coadd
                  ;
                  raw_flag:  0, $                   ; 0 = no image in raw buffer; 1 = image present in raw buffer
                  raw_image: dblarr(3000,3000), $   ; space for reading a raw Mimir image - note need to use REFORM
                  ncol_size:    1500L,  $            ; actual size of space in columns
                  nrow_size:    1500L,  $            ; same for columns
                  ncol_raw: 0L,      $               ; number of actual columns in raw image
                  nrow_raw: 0L,   $
                  raw_FITS_header: strarr(300), $   ; holder for FITS header
                  raw_FITS_lines: 0, $              ; number of lines in FITS header
                  ;
                  input_dir: '', $                  ; directory for input images
                  output_dir: '', $                 ; directory for output images
                  suffix: '_Ar', $                  ; file name suffix to insert after the file name and before the ".FITS"
                  ;
                  result_flag:  0,   $              ; 0 = no image in result buffer; 1 = image is present
                  result_image:dblarr(3000,3000), $ ; central portion of tall slit for F5 modes
                  ncol_result: 0L,     $             ; number of columns in trimmed image
                  nrow_result: 0L,     $
                  trim_col: 0L, $                ; number of columns trimmed, including column 0
                  trim_row: 0L, $                ; number of rows trimmed, including row 0
                  result_FITS_header: strarr(300), $
                  result_FITS_lines: 0, $
                  result_file_root: '',     $       ; file name of first image in coadd stack
                  ;
                  ;  difference image (usually between result image and MFM version to enhance cal lines)
                  ;
                  dif_image: dblarr(3000,3000),$    ; the image
                  ncol_dif: 0L, $                ; number of columns
                  nrow_dif: 0L, $                ; and rows
                  ;
                  base_window_no:   -1L,   $        ; window number of the plot region of the base frame
                  base_xsize: -1L, $             ; x size of window on open
                  base_ysize: -1L, $             ; y size of window on open
                  ;
                  S4_top_win_no: -1L,   $           ; window number of top window for S4 pop-up
                  S4_middle_win_no: -1L,    $       ;   "             middle
                  S4_bottom_win_no: -1L,     $      ;   "             bottom
                  S4_widget_ID: -1L, $              ; base widget number of the pop up window
                  ;
                  Argon: Argon, $                   ; Argon line list structure
                  OH: OH, $                         ; OH line list structure
                  LM: LM_sky, $						; LM sky feature list
                  ;
                  disp_0: -1.0d0, $                 ; nominal dispersion equation, offset term [in microns]
                  disp_1: -1.0d0, $                 ;             "                linear
                  disp_2: -1.0d0,  $                ;             "                parabolic
                  ;
                  nch: -1L, $                       ; number of channels in the spectrum
                  x_spec: dblarr(3000), $           ; x values of spectrum
                  y_spec: dblarr(3000), $           ; y values
                  ;
                  nzones: -1, $                     ; number of row zones
                  zone_size: 30, $                  ; row zone size
                  new_zone_size: 0, $               ; the row zone size from the text box buffer
                  row_zones: intarr(3000,2),  $      ; starting, stopping row numbers for up to 3000 zones
                  ;
                  ; 2D dispersion coeffs
                  ;
                  disp_2D: dblarr(3,3),  $         ; dual axis parabola to give wavelength as function of col, row
                  ;
                  oversample: 2.0, $               ; factor by which to oversample the input spectrum in making the output spectrum
                  ;
                  ;   nominal dispersion equations
                  ;
                  dispersion_library: {jhk_jhk_dispersion: 	jhk_jhk_dispersion, $
                  			 jhk_hk_dispersion: 			jhk_hk_dispersion, $
                             jhk_k_dispersion: 				jhk_k_dispersion, $
                             lm_1st_dispersion: 			lm_1st_dispersion, $
                      ;       lm_2nd_dispersion: lm_2nd_dispersion, $
                             lm_3rd_dispersion: 			lm_3rd_dispersion, $
                             lm_4th_dispersion: 			lm_4th_dispersion, $
                             sed_dispersion: 				sed_dispersion, $
                             lm_2nd_k17_dispersion: 		lm_2nd_k17_dispersion, $
                             jhk_izyj_1st_dispersion: 		jhk_izyj_1st_dispersion, $
                             jhk_izyj_2nd_dispersion: 		jhk_izyj_2nd_dispersion, $
                             jhk_h17_dispersion:  			jhk_h17_dispersion, $
                             jhk_iz17_2nd_dispersion: 		jhk_iz17_2nd_dispersion, $
                             bvr_vph_dispersion: 			bvr_vph_dispersion} $
                  }
    ;
    ;print, "creating structure for widget ID = ",wWidget
    ;

    WIDGET_CONTROL, wWidget, SET_UVALUE=WCT_struct, /NO_COPY
end
;-----------------------------------------------------------------
; Notify Realize Callback Procedure.
; Argument:
;   wWidget - ID number of specific widget.
;
;
;   Retrieve the IDs of other widgets in the widget hierarchy using
;       id=widget_info(Event.top, FIND_BY_UNAME=name)

;-----------------------------------------------------------------
pro WCT_realize, wWidget
    ;
    ;   set all default buttons, modes
    ;
    ;print, "in realize, widget id = ",wWidget
    ;
    WIDGET_CONTROL, wWidget, GET_UVALUE=WCT_struct, /NO_COPY
    ;
    WCT_struct.instrument = 0
    instrument = 0
    WCT_struct.camera = 0
    camera = 0
    WCT_struct.slit = 1
    slit = 1
    WCT_struct.object = 0
    object = 0
    WCT_struct.spec_mode = 1
    spec_mode = 1
    WCT_struct.cal_mode = 0
    cal_mode = 0
    WCT_struct.file_mode = 0
    ;
   	TLB_Title = WCT_struct.version_string
    tlb_wid = WIDGET_INFO( wWidget, FIND_BY_UNAME='WID_BASE_0')
    WIDGET_CONTROL, tlb_wid, TLB_SET_TITLE=TLB_Title
	;
    ;   size of base window
    ;
    window_id = WIDGET_INFO( wWidget, FIND_BY_UNAME="base_plot")
    window_geo = WIDGET_INFO( window_id, /GEOMETRY)
;    WCT_struct.base_xsize = window_geo.draw_xsize
;    WCT_struct.base_ysize = window_geo.draw_ysize
    ;
    WCT_struct.base_xsize = window_geo.scr_xsize
    WCT_struct.base_ysize = window_geo.scr_ysize
    ;
    ;help, window_geo, /struct
    ;print, "base plot window size = ",WCT_struct.base_xsize, WCT_struct.base_ysize
    ;
    ;     interpolation/fitting limits
    ;
    wave_limits = WCT_struct.wave_limits
    wave_limits[0,0:1] = [1.18d0, 2.28d0]      	; JHK-Spec limits
    wave_limits[1,0:1] = [1.9d0, 2.5d0]        	; K-Spec limits
    wave_limits[2,0:1] = [1.17d0, 1.36d0]      	; J-spec 3rd order limits
    wave_limits[3,0:1] = [1.17d0, 1.36d0]	   	; J-spec 4th order
    wave_limits[4,0:1] = [2.8d0, 5.5d0]			; LM Spec limits
    wave_limits[5,0:1] = [1.40d0, 2.5d0]		; HK-Spec limits
    wave_limits[6,0:1] = [1.80d0,2.35d0]		; K17-Spec limits
    wave_limits[7,0:1] = [1.02d0,1.38d0]		; IZYJ-Spec 1st order
    wave_limits[8,0:1] = [0.80d0,1.274d0]		; IZYJ-Spec 2nd order
    wave_limits[9,0:1] = [1.517d0, 1.978d0]     ; H17-Spec
    wave_limits[10,0:1] = [0.8d0,1.0d0]			; IZ17-Spec
    wave_limits[11,0:1] = [0.48d0, 0.78d0]		; BVR VPH limits
    WCT_struct.wave_limits= wave_limits
    ;
    ;   show selected instrument tab
    ;
    inst_tab = WIDGET_INFO( wWidget, FIND_BY_UNAME="Instrument_tab")
    WIDGET_CONTROL, inst_tab, SET_TAB_CURRENT=WCT_struct.instrument
    ;
    for instrument = 0, 1 do begin
        ;
        ;  show selected camera radio button
        ;
        WCT_S1_Set_Camera_buttons, wWidget, instrument, camera
        ;
        ;  show selected spectroscopy mode
        ;
        WCT_S1_Set_Spec_buttons, wWidget, instrument, spec_mode
        ;
        ;  slit mode
        ;
        WCT_S1_Set_Slit_buttons, wWidget, instrument, slit
        ;
        ;  object mode
        ;
        WCT_S1_Set_Object_buttons, wWidget, instrument, object
        ;
        ;  calibration mode - start with Argon for all instruments
        ;
    endfor
    ;
    argon_id = WIDGET_INFO( wWidget, FIND_BY_UNAME="S2_Argon_On")
    WIDGET_CONTROL, argon_id, SET_BUTTON=0
    oh_id = WIDGET_INFO( wWidget, FIND_BY_UNAME="S2_OH_On")
    WIDGET_CONTROL, oh_id, SET_BUTTON=0
    lm_id = WIDGET_INFO( wWidget, FIND_BY_UNAME = "S2_LM_On")
    WIDGET_CONTROL, lm_id, SET_BUTTON=0
    ;
    case WCT_struct.cal_mode of
        0: WIDGET_CONTROL, argon_id, SET_BUTTON=1
        1: WIDGET_CONTROL, oh_id, SET_BUTTON=1
        2: WIDGET_CONTROL, lm_id, SET_BUTTON=1
    endcase
    ;
    WIDGET_CONTROL, wWidget, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    ;   fudge up structure to fake an Event structure
    ;
    fake_event = {top: wWidget, id: wWidget, handle: wWidget}
    WCT_save_step, fake_event, 0
    WCT_show_status, fake_event
end
;------------------------------------- Full Exit --------------
pro base_exit, Event
    ;
    if(S4_PU_Exists(Event)) then WCT_S4_PU_Kill, Event
    ;
    top = Event.top
    WIDGET_CONTROL, top, /DESTROY
end
;------------------------------------- Print Summary ----------------
pro wsummary_print, wWidget
    ;
    ;  write text to the summary window
    ;
    summary_ID = WIDGET_INFO( wWidget, FIND_BY_UNAME="Summary_text_window")
    ;file_in = dialog_pickfile()
    ;print, "File name = "
    file_in = FILE_SEARCH('WCT_summary.txt',/FULLY_QUALIFY_PATH)
    ;
    ;print, file_in
    openr, lu_in, file_in, /GET_LUN
    ist = 0
    WHILE ~EOF(lu_in) do begin
       line = ''
       readf, lu_in, line
       WIDGET_CONTROL, summary_ID, SET_VALUE=line, /APPEND
       ist++
    ENDWHILE
    FREE_LUN, lu_in
    ;print, "read in ",ist," lines"

end

;----------------------------------------- Move to Next Tab ------------
pro New_Tab, Event
    ;
    ;  changed to another tab. Based on which one, refresh that tab's information, if
    ;  it exists
    ;
    tab_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="Step_tab")
    tab_no = WIDGET_INFO( tab_id, /TAB_CURRENT)
    ;print,"moved to tab number ",tab_no
    ;print, "trying to load structure at widget ID = ",Event.top

    ;print, "struct of WCT_struct:"
    ;help, WCT_struct
    ;help, WCT_struct, /structures
    ;
    case tab_no of
       ;
       ; summary - do nothing
       0: begin
            WCT_save_step, Event, 0
            WCT_show_status, Event
          end
       ;
       ; select spec mode - if mode selected, echo, else clear all
       ;
       1:  begin         ;     start by clearing all
               S1_New_Tab, Event
               ;
               WCT_save_step, Event, 1
               WCT_show_status, Event
           end
       ;
       ; select cal mode
       ;
       2: begin        ;  start by clearing all
             WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
             Ar_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S2_Argon_On")
             WIDGET_CONTROL, Ar_ID, SET_BUTTON=0
             ;
             OH_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S2_OH_On")
             WIDGET_CONTROL, OH_ID, SET_BUTTON=0
             ;
             LM_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S2_LM_On")
             WIDGET_CONTROL, LM_ID, SET_BUTTON=0
             ;
             case WCT_struct.cal_mode of
              -1:
              0: WIDGET_CONTROL, Ar_ID, SET_BUTTON=1
              1: WIDGET_CONTROL, OH_ID, SET_BUTTON=1
              2: WIDGET_CONTROL, LM_ID, SET_BUTTON=1
             endcase
             WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
             ;
             WCT_save_step, Event, 2
             WCT_show_status, Event
         end

       ;
       ; set up and display cal images
       ;
       3: begin           ;  clear all
             WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
             Input_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S3_Input_dir_text")
             if(WCT_struct.input_dir ne '') then begin
              WIDGET_CONTROL, Input_ID, SET_VALUE=WCT_struct.input_dir
             endif else begin
              WIDGET_CONTROL, Input_ID, SET_VALUE=''
             endelse
             ;
             Output_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S3_Output_dir_text")
             if(WCT_struct.output_dir ne '') then begin
              WIDGET_CONTROL, Output_ID, SET_VALUE=WCT_struct.output_dir
             endif else begin
              WIDGET_CONTROL, Output_ID, SET_VALUE=''
             endelse
             ;
             Single_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S3_Single_On")
             WIDGET_CONTROL, Single_ID, SET_BUTTON=0
             Multiple_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S3_Multiple_On")
             WIDGET_CONTROL, Multiple_ID, SET_BUTTON=0
             ;
             case WCT_struct.file_mode of
                -1:
                0: WIDGET_CONTROL, Single_ID, SET_BUTTON=1
                else: WIDGET_CONTROL, Multiple_ID, SET_BUTTON=1
             endcase
             WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
             ;
             WCT_save_step, Event, 3
             WCT_show_status, Event
           end
         ;
         ;  find slit offset, dispersion equations
         ;
         4: begin
              if(S4_PU_Exists(Event)) then WCT_S4_PU_Kill, Event
              ;
              WCT_save_step, Event, 4
              WCT_show_status, Event
            end
         ;
         5: begin
              if(S4_PU_Exists(Event)) then WCT_S4_PU_Kill, Event
              Zone_size_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_Zone_size_text_box")
              WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
              ;
              ;	special case of fat-eared slit in Mimir
              ;
              if(WCT_struct.instrument eq 0 and WCT_struct.slit eq 4) then begin
              	WCT_struct.zone_size = 39
              endif
              ;
              zone_size = WCT_struct.zone_size
              WIDGET_CONTROL, Zone_size_id, SET_VALUE=STRING(format='(i3)',zone_size)
              WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
              ;
              ;   call the calculator
              ;
              S5_Zone_size_text_box_input, Event
              ;
              WCT_save_step, Event, 5
              WCT_show_status, Event
            end
         6: begin
              if(S4_PU_Exists(Event)) then WCT_S4_PU_Kill, Event
              ;
              WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
              	;
              	;	display current input, output directories, if already selected
              	;
              	in_dir = WCT_struct.input_dir
              	if(in_dir ne '') then begin
              		text_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S6_Input_Directory_text_box")
        			WIDGET_CONTROL, text_ID, SET_VALUE=in_dir
        		endif
        		out_dir = WCT_struct.output_dir
        		if(out_dir ne '') then begin
        			text_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S6_Output_Directory_text_box")
        			WIDGET_CONTROL, text_ID, SET_VALUE=out_dir
        		endif
        		;
        		;	list the suffix in its box
        		;
                  text_string = WCT_struct.suffix
                  ;
                  text_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S6_Suffix_text_box")
               ;
               WIDGET_CONTROL, text_box_id, SET_VALUE=text_string
             WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
             ;
              WCT_save_step, Event, 6
              WCT_show_status, Event
              ;
            end
         7: begin
         	  base_exit, Event
            end
         8: begin
            end
         9: begin
            end
         else: begin
              end
    endcase
end

;--------------------------------- Step 0 --------------------
pro S0_Done, Event
    WCT_print_text, Event, "Closing Tab 0, Opening Tab 1"
    ;
    ;   open Tab 1 (Step 1)
    ;
    tab_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="Step_tab")
    WIDGET_CONTROL, tab_ID, SET_TAB_CURRENT=1
    New_Tab, Event
end
;--------------------------------- Step 1 ---------------------
pro S1_JHK_Spec_On, Event
    WCT_S1_Change_Spec_Mode, Event, 0, "Selecting JHK-Spec Mode"
    WCT_show_status, Event
end
;---------------
pro S1_K_Spec_On, Event
	;
	;	if f17 not selected, is normal K-Spec
	;
	WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
	camera = WCT_struct.camera
	WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
	if(camera eq 0) then begin
    	WCT_S1_Change_Spec_Mode, Event, 1, "Selecting K-Spec Mode"
    endif else begin
    ;
    ;	otherwise, is K17-Spec
    ;
    	WCT_S1_Change_Spec_Mode, Event, 6, "Selecting K17-Spec Mode"
    endelse
    WCT_show_status, Event
end
;------------
pro S1_J_Spec_3rd_On, Event
    WCT_S1_Change_Spec_Mode, Event, 2, "Selecting J-Spec (3rd Order) Mode"
    WCT_show_status, Event
end
;------------
pro S1_J_Spec_4th_On, Event
    WCT_S1_Change_Spec_Mode, Event, 3, "Selecting J-Spec (4th Order) Mode"
    WCT_show_status, Event
end
;------------
pro S1_LM_Spec_On, Event
    WCT_S1_Change_Spec_Mode, Event, 4, "Selecting LM-Spec (1st Order) Mode"
    WCT_show_status, Event
end
;------------
pro S1_HK_Spec_On, Event
    WCT_S1_Change_Spec_Mode, Event, 5, "Selecting HK-Spec Mode"
    WCT_show_status, Event
end
pro S1_K17_Spec_On, Event
	WCT_S1_Change_Spec_Mode, Event, 6, "Selecting K17-Spec Mode"
	WCT_show_status, Event
end
pro S1_IZYJ_1_Spec_On, Event
	WCT_S1_Change_Spec_Mode, Event, 7, "Selecting IZYJ-Spec 1st Order Mode"
	WCT_show_status, Event
end
pro S1_IZYJ_2_Spec_On, Event
	WCT_S1_Change_Spec_Mode, Event, 8, "Selecting IZYJ-Spec 2nd Order Mode"
	WCT_show_status, Event
end
pro S1_H17_Spec_On, Event
   	WCT_S1_Change_Spec_Mode, Event, 9, "Selecting H17-Spec Mode"
   	WCT_show_status, Event
end
pro S1_IZ17_Spec_On, Event
	WCT_S1_Change_Spec_Mode, Event, 10, "Selecting IZ17-Spec Mode"
	WCT_show_status, Event
end
;-----------------
pro S1_PRISM_VPH_On, Event
    WCT_print_text, Event, "VPH-Spec Not Yet Implemented"
    ;
    ;   clear VPH button
    ;
    VPH_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S1_PRISM_VPH_On")
    WIDGET_CONTROL, VPH_ID, SET_BUTTON=0
end
;------------
pro S1_F5_On, Event
    WCT_S1_Change_Camera_Mode, Event, 0, "Selecting F5 Camera mode"
    WCT_show_status, Event
end
;------------
pro S1_F17_On, Event
    WCT_S1_Change_Camera_Mode, Event, 1, "Selecting F17 Camera mode"
    WCT_show_status, Event
end
;------------
pro S1_PRISM_Single_On, Event
    WCT_S1_Change_Object_Mode, Event, 0, "Selecting Single Object mode"
end
;------------
pro S1_PRISM_Multiple_On, Event
    WCT_S1_Change_Object_Mode, Event, 1, "Selecting Multiple Object mode"
end
;------------
pro S1_PRISM_Short_Slit_On, Event
    WCT_S1_Change_Slit_Mode, Event, 0, "Selecting PRISM Short Slit mode"
end
;------------
pro S1_PRISM_Tall_Slit_On, Event
    WCT_S1_Change_Slit_Mode, Event, 1, "Selecting PRISM Tall Slit mode"
end
;------------------------
pro S1_Slit_Select_1, Event
    WCT_S1_Change_Slit_Mode, Event, 1, "Selecting Mimir Tall-Narrow Slit mode"
end
;------------
pro S1_Slit_Select_4, Event
    WCT_S1_Change_Slit_Mode, Event, 4, "Selecting Mimir Fat-Eared Slit mode"
end
;------------
pro S1_Slit_Select_11, Event
    WCT_S1_Change_Slit_Mode, Event, 11, "Selecting Mimir Short-Narrow Slit mode"
end
;------------
pro S1_New_Tab, Event
    ;
    ;  this is called when ever the subtab base switches from Mimir to PRISM
    ;
    tab_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="Instrument_tab")
    tab_no = WIDGET_INFO( tab_id, /TAB_CURRENT)
    case tab_no of
        0: begin
               ;     Mimir
               ;
               instrument = 0
               WCT_print_text, Event, "Selecting Mimir instrument"
           end
        1: begin
               instrument = 1
               WCT_print_text, Event, "Selecting PRISM instrument"
           end
    endcase
    ;
    ;  reload all the current stuff
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    WCT_struct.instrument = instrument
    spec_mode = WCT_struct.spec_mode
    camera = WCT_struct.camera
    slit = WCT_struct.slit
    object = WCT_struct.object
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    Reset_Mask_On_Mode_Change, Event
    ;
    WCT_S1_Set_Spec_buttons, Event.top, instrument, spec_mode
    WCT_S1_Set_Camera_buttons, Event.top, instrument, camera
    WCT_S1_Set_Slit_buttons, Event.top, instrument, slit
    WCT_S1_Set_Object_buttons, Event.top, instrument, object
    WCT_show_status, Event
end
;--------------------
pro S1_Done, Event
    WCT_print_text, Event, "Closing Tab 1, Opening Tab 2"
    ;
    ;   open Tab 2 (Step 2)
    ;
    tab_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="Step_tab")
    WIDGET_CONTROL, tab_ID, SET_TAB_CURRENT=2
    New_Tab, Event
end
;---------------------------- Step 2 ------------------------
pro S2_Argon_On, Event
    WCT_print_text, Event, "Argon Lamp Calibration Selected"
    ;
    ;  set this button, clearing the other
    ;
    Ar_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S2_Argon_On")
    WIDGET_CONTROL, Ar_ID, SET_BUTTON=1
    ;
    ;   clear other buttons
    ;
    OH_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S2_OH_On")
    WIDGET_CONTROL, OH_ID, SET_BUTTON=0
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    WCT_struct.cal_mode = 0
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    WCT_show_status, Event
end

;------------------
pro S2_OH_On, Event
    WCT_print_text, Event, "OH Night Sky Calibration Selected"
    ;
    ;  set this button, clearing the other
    ;
    Ar_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S2_Argon_On")
    WIDGET_CONTROL, Ar_ID, SET_BUTTON=0
    ;
    ;   clear other buttons
    ;
    OH_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S2_OH_On")
    WIDGET_CONTROL, OH_ID, SET_BUTTON=1
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    WCT_struct.cal_mode = 1
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    WCT_show_status, Event
end

;------------------
pro S2_LM_On, Event
    WCT_print_text, Event, "LM Night Sky Feature Calibration Selected"
    ;
    ;   clear other buttons
    ;
    Ar_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S2_Argon_On")
    WIDGET_CONTROL, Ar_ID, SET_BUTTON=0
    ;
    OH_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S2_OH_On")
    WIDGET_CONTROL, OH_ID, SET_BUTTON=0
    ;
    ;
    ;  set this button, clearing the other
    ;
    LM_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S2_LM_On")
    WIDGET_CONTROL, LM_ID, SET_BUTTON=1
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    WCT_struct.cal_mode = 2
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    WCT_show_status, Event
end
;-------------------
pro S2_Done, Event
    WCT_print_text, Event, "Closing Tab 2, Opening Tab 3"
    ;
    ;   open Tab 3 (Step 3)
    ;
    tab_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="Step_tab")
    WIDGET_CONTROL, tab_ID, SET_TAB_CURRENT=3
    New_Tab, Event
end

;---------------------------- Step 3 -----------------------
;
;---------------- Part 1 - Directories -------
;
pro S3_Input_Directory, Event
    WCT_print_text, Event, "Setting Input Directory"
    dir_in = dialog_pickfile(TItle="Select Input Directory",/directory)
    if(dir_in ne '') then begin
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        WCT_struct.input_dir = dir_in
        ;print, "Input Directory = ",dir_in
        WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
        ;
        text_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S3_Input_dir_text")
        ;print,"input text widget id = ",text_ID
        WIDGET_CONTROL, text_ID, SET_VALUE=dir_in
    endif else begin
       WCT_print_text, Event, "Cancel Operation Detected. Continuing"
    endelse
end

;
pro S3_Output_Directory, Event
    WCT_print_text, Event, "Setting Output Directory"
    dir_out = dialog_pickfile(Title="Select Output Directory",/directory)
    if(dir_out ne '') then begin
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        WCT_struct.output_dir = dir_out
        ;print, "Output Directory = ",dir_out
        WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
        ;
        text_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S3_Output_dir_text")
        ;print,"output text widget id = ",text_ID
        WIDGET_CONTROL, text_ID, SET_VALUE=dir_out
    endif else begin
       WCT_print_text, Event, "Cancel Operation Detected. Continuing"
    endelse
end

pro S3_Done, Event
    WCT_print_text, Event, "Closing Tab 3, Opening Tab 4"
    ;
    ;   open Tab 4 (Step 4)
    ;
    tab_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="Step_tab")
    WIDGET_CONTROL, tab_ID, SET_TAB_CURRENT=4
    New_Tab, Event
end
;
;---------------- Part 2 - Files -------------
;
pro S3_Single_On, Event
    WCT_print_text, Event, "Single Image Option Selected"
    ;
    ;  set this button, clearing the other
    ;
    S_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S3_Single_On")
    WIDGET_CONTROL, S_ID, SET_BUTTON=1
    ;
    ;   clear other buttons
    ;
    M_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S3_Multiple_On")
    WIDGET_CONTROL, M_ID, SET_BUTTON=0
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    WCT_struct.file_mode = 0
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
end

;-------------------------
pro S3_Multiple_On, Event
    WCT_print_text, Event, "Multiple Images Option Selected"
    ;
    ;  set this button, clearing the other
    ;
    S_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S3_Single_On")
    WIDGET_CONTROL, S_ID, SET_BUTTON=0
    ;
    ;   clear other buttons
    ;
    M_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S3_Multiple_On")
    WIDGET_CONTROL, M_ID, SET_BUTTON=1
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    WCT_struct.file_mode = 1
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
end
;---------------------
pro S3_Create_Input_On, Event
    ;
    ;
    ;  start with guess of old input directory
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    input_dir = WCT_struct.input_dir
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    ;  create the input directory and set path to it
    ;
    WCT_print_text, Event, "Use file chooser to create new directory"
    ;
    dummy = 'dummy'
    ;
    if(input_dir ne '') then begin
        file_dummy = dialog_pickfile(title="Create New Input Directory", $
           FILE=dummy, PATH=input_dir)
    endif else begin
        file_dummy = dialog_pickfile(title="Create New Input Directory", $
           FILE=dummy)
    endelse
    ;
    if(file_dummy ne '') then begin
        ;
        ;  trim pathname to get just the directory path
        ;
        input_dir = FILE_DIRNAME(file_dummy, /MARK_DIRECTORY)
        ;
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        WCT_struct.input_dir = input_dir
        FILE_MKDIR, input_dir
        WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
        ;
        text_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S3_Input_dir_text")
        WIDGET_CONTROL, text_ID, SET_VALUE=input_dir
    endif else begin
       WCT_print_text, Event, "Cancel Operation Detected. Continuing"
    endelse
end

;----------------------
pro S3_Create_Output_On, Event
    ;
    ;  start with guess of old output directory
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    output_dir = WCT_struct.output_dir
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    ;  create the output directory and set path to it
    ;
    WCT_print_text, Event, "Use file chooser to create new directory"
    ;
    dummy = 'dummy'
    ;
    if(output_dir ne '') then begin
        file_dummy = dialog_pickfile(title="Create New Output Directory", $
           FILE=dummy, PATH=output_dir)
    endif else begin
        file_dummy = dialog_pickfile(title="Create New Output Directory", $
           FILE=dummy)
    endelse
    ;
    if(file_dummy ne '') then begin
        ;
        ;  trim pathname to get just the directory path
        ;
        output_dir = FILE_DIRNAME(file_dummy, /MARK_DIRECTORY)
        ;
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        WCT_struct.output_dir = output_dir
        FILE_MKDIR, output_dir
        WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
        ;
        text_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S3_Output_dir_text")
        WIDGET_CONTROL, text_ID, SET_VALUE=output_dir
    endif else begin
       WCT_print_text, Event, "Cancel Operation Detected. Continuing"
    endelse
end
;
;---------------- Part 3 - Processing --------
;
pro S3_fetch_on, Event
    ;
    ;  test for file mode
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    ;
    ;   examine current file mode
    ;
    file_mode = WCT_struct.file_mode
    ;
    case file_mode of
        -1:    begin          ; not yet set, so can't fetch
                  WCT_print_text, Event, "ERROR - Must select file mode first"
                end
        ;
         0: begin          ; single
             ;
             ;   if directory set, try to fetch from there
             ;
             if(WCT_struct.input_dir ne '') then begin
                   calibration_file = dialog_pickfile(title="Select Calibration Image", PATH=WCT_struct.input_dir)
               endif else begin
                   calibration_file = dialog_pickfile(title="Select Calibration Image")
               endelse
               ;
               if(calibration_file ne '') then begin
                   ;
                   WCT_print_text, Event, STRING(format='("Read In Cal Image = ",a)',calibration_file)
                   ;
                   ;   read image, update structure
                   ;
                   WCT_struct.file_name = calibration_file
                   WCT_struct.coadd = 0
                   image = readfits(calibration_file,header,/SILENT)
                   ind_nan = WHERE(~FINITE(image), n_nan)
                   if(n_nan gt 0) then image[ind_nan] = -1.0e6
                   i_size = SIZE(image)
                   WCT_struct.ncol_raw = i_size[1]
                   WCT_struct.nrow_raw = i_size[2]
                   ;
                   ;    pack input array to fit in lower left of storage array
                   ;
                   WCT_struct.raw_image[0:i_size[1]-1,0:i_size[2]-1] = double(image)
                   WCT_struct.raw_image[i_size[1]:*,i_size[2]:*] = 0.0d0
                   ;
                   nlines = N_ELEMENTS(header)
                   WCT_struct.raw_FITS_header[0:nlines-1] = header
                   WCT_struct.raw_FITS_header[nlines:*] = ''
                   WCT_struct.raw_FITS_lines = nlines
                   ;
                   ;    set raw flag
                   ;
                   WCT_struct.raw_flag = 1
                   ;
                   ;    but, if the CAL-MODE or SPECMODE
                   ;       keywords are present, move the image to the RESULT buffer
                   ;
                   ;    also, if TRIM-COL or TRIM-ROW keywords are present, move to RESULT buffer
                   ;
                   ;       and get the dispersion equation????
                   ;
                   ical = FXPAR(header, "SPECMODE", COUNT=icount)

              ;
                 trim_col = FXPAR(header, "TRIM-COL", count=jcount)
                    if(jcount gt 0) then WCT_struct.trim_col = trim_col
                    trim_row = FXPAR(header, "TRIM-ROW", count=kcount)
                    if(kcount gt 0) then WCT_struct.trim_row = trim_row

                   if(icount+jcount+kcount ne 0) then begin
                        WCT_print_text, Event, "Read Image already partially processed"
                        ;print, "Read Image Already Wavelength Calibrated"
                        ;
                        ;  already partially processed data, move to result buffer
                        ;
                        WCT_struct.result_image = WCT_struct.raw_image
                        WCT_struct.ncol_result = WCT_struct.ncol_raw
                        WCT_struct.nrow_result = WCT_struct.nrow_raw
                        ;
                        WCT_struct.nrow_raw = 0
                        WCT_struct.ncol_raw = 0
                        ;
                        WCT_struct.raw_flag = 0
                        WCT_struct.result_flag = 1
                        ;
                        WCT_struct.result_FITS_header = WCT_struct.raw_FITS_header
                        WCT_struct.result_FITS_lines = WCT_struct.raw_FITS_lines
                        ;
                        WCT_struct.raw_FITS_lines = 0
                        WCT_struct.raw_FITS_header[*] = ''
                        ;

                   endif
                   ;
                   ;    plot it, first saving structure
                   ;
                   WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
                   S3_Display_On, Event
                   WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
               endif else begin
                      WCT_print_text, Event, "Cancel Recognized. No file read"
               endelse
           end
         ;
         else: begin      ; multiple
                ;
                ;       if directory set, try to fetch from there
                ;
                if(WCT_struct.input_dir ne '') then begin
                     calibration_files = dialog_pickfile(title="Select Calibration Images", /Multiple_Files, PATH=WCT_struct.input_dir)
                 endif else begin
                     calibration_files = dialog_pickfile(title="Select Calibration Images", /Multiple_Files)
                 endelse
                 ;
                 ;    how many?
                 ;
                 nfiles = N_ELEMENTS(calibration_files)
                 if(nfiles gt 1) then begin
                     ;
                     WCT_print_text, Event, STRING(format='("Selected ",i4," Calibration Images")',nfiles)
                     ;
                     ;   read images, update structure
                     ;
                     WCT_struct.file_name = ''
                     WCT_struct.coadd = nfiles
                     WCT_struct.file_mode = nfiles
                     WCT_struct.nfiles = nfiles
                     ;
                     WCT_struct.files_multi[0:nfiles-1] = calibration_files    ; note that we don't read these in just yet
                     WCT_struct.files_multi[nfiles:*] = ''
                     ;
                     WCT_struct.result_file_root = calibration_files[0]     ; first image name
                     ;
                     ;  clear raw flag
                     ;
                     WCT_struct.raw_flag = 0
                     WCT_struct.result_flag = 1
                     ;
                     ;  start coadding
                     ;
                     WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
                     S3_Coadd_On, Event
                     WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
                     ;
                 endif else begin
                   WCT_print_text, Event, "Cancel Recognized. No files read."
                 endelse
            end
    endcase
    ;
    ;  save structure
    ;
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
end

;--------------------
pro S3_Display_On, Event
    ;
    ;  first check that a valid image is in structure somewhere
    ;
    ;     raw has priority, result next
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_Struct, /NO_COPY
    ;
    idebug = WCT_struct.debug_flag
    ;
    if(WCT_Struct.raw_flag ne 0 or WCT_struct.result_flag ne 0) then begin
        if(WCT_struct.base_window_no eq -1L) then begin
            window_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="base_plot")
            WIDGET_CONTROL, window_ID, GET_VALUE=window_no
            WCT_struct.base_window_no = window_no
        endif
        ;
       window_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="base_plot")
       window_no = WCT_struct.base_window_no
       ;
       ;    recapture image from storage array
       ;
       if(WCT_struct.raw_flag ne 0) then begin
           ;
           ;    fetch just enough data
           ;
           ncol_plot = WCT_struct.ncol_raw
           nrow_plot = WCT_struct.nrow_raw
           image = WCT_struct.raw_image[0:ncol_plot-1,0:nrow_plot-1]
       endif else begin
           ;
           ;   result, if no raw
           ;
           ncol_plot = WCT_struct.ncol_result
           nrow_plot = WCT_struct.nrow_result
           image = WCT_struct.result_image[0:ncol_plot-1,0:nrow_plot-1]
       endelse
       ;
       ;    reset the drawing size, if necessary
       ;
       WIDGET_CONTROL, window_ID, DRAW_XSIZE=ncol_plot, DRAW_YSIZE=nrow_plot
       ;
       ;    display
       ;
       z = WCT_display_image(image, window_no)
       ;
    endif
    ;
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_Struct, /NO_COPY
end

;-----------------------
pro S3_Mask_On, Event
    ;
    ;  Mask/Trim image
    ;
    ;  NB: need to add a popup to select either F5 or custom masking
    ;
    ;
    WCT_print_text, Event, "Applying Image Masking"
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    ;
    idebug = WCT_struct.debug_flag
    ;
    ;   Mask based on instrument and camera modes
    ;
    instrument = WCT_struct.instrument
    camera = WCT_struct.camera
    spec_mode = WCT_struct.spec_mode
    slit = WCT_struct.slit
    ;
    xl = WCT_struct.trim_col
    yl = WCT_struct.trim_row
    xu = WCT_struct.ncol_result + xl - 1
    yu = WCT_struct.nrow_result + yl - 1
    ;
    xcol = 1024
    xrow = 1026
    ;
    ;------------------------------------------------------------
    ;      masking information - some of this could
    ;          be collected from the user, or hardwired
    ;          --- a work still in process ----
;    case instrument of
;           0:    begin                 ; Mimir
;              case camera of
;                  0: begin          ; F/5
;                       xcol = 1024   ; original sizes
;                       xrow = 1026
;                       ;             ; mask limits
;                       case slit of
;                       		1: begin 	;	tall narrow
;                         		yl = 260
;                         		yu = 759
;                         	   end
;                         	4: begin	; 	fat-eared
;                         		yl = 201
;                         		yu = 813	; of height 40 pixels each
;                         	   end
;                         	11: begin	; 	short narrow
;                         		 yl = 360	; TBD
;                         		 yu = 660	; TBD
;                         		end
;                        endcase
;                         xl = 0
;                         xu = 1023     ; no x masking
;                    end
;                  1: begin          ; F/17
;                       xcol = 1024   ; original sizes
;                       xrow = 1026
;                       yl = 0
;                         yu = 1025     ; no y masking
;                         xl = 0
;                         xu = 1023     ; no x masking
;                    end
;                    ;
;                    ;   N.B. this will need extreme rework for sub-framing, esp LM images
;                    ;
;              endcase
;           end
;          1: begin                    ; PRISM
;               end                    ;   TBD
;    endcase
;    ;
;    WCT_struct.nrow_result = yu - yl + 1
;    WCT_struct.trim_row = yl
;    WCT_struct.trim_col = xl
;    WCT_struct.ncol_result = xu - xl + 1
    ;
    ;------------------------------------------------------------
    ;
    ;  check:
    ;;
    ;     this check is a kludge - it looks only to see if the image size matches
    ;      the size of a raw image. If so, masking is permitted, if not masking
    ;      is not permitted. This may not be a valid decision tree...
    ;
    ;       first in raw, then in result
    ;       for image of 1024x1026 size
    ;
    if(WCT_struct.raw_flag ne 0 or WCT_struct.result_flag ne 0) then begin
       ;
       ; check raw first
       ;
       if(WCT_struct.raw_flag ne 0) then begin
             if(WCT_struct.ncol_raw eq xcol and WCT_struct.nrow_raw eq xrow) then begin
                 image = WCT_struct.raw_image[0:xcol-1,0:xrow-1]
                 FITS_header = WCT_struct.raw_FITS_header[0:WCT_struct.raw_FITS_lines-1]
                 file_name = WCT_struct.file_name
                 iskip = 0
             endif else begin
                 iskip = 1
                 WCT_print_text, Event, "ERROR - Image in Raw Buffer wrong size"
             endelse
       endif else begin
         ;
         ;    then result
         ;
         if(WCT_struct.result_flag ne 0) then begin
              if(WCT_struct.ncol_result eq xcol and WCT_struct.nrow_result eq xrow) then begin
                  image = WCT_struct.result_image[0:xcol-1, 0:xrow-1]
                  FITS_header = WCT_struct.result_FITS_header[0:WCT_struct.result_FITS_lines-1]
                  file_name = WCT_struct.result_file_root
                  iskip = 0
              endif else begin
                  WCT_print_text, Event, "ERROR - Image in Result Buffer already masked"
                  iskip = 1
              endelse
         endif
         ;
       endelse
       ;
       if(iskip eq 0) then begin
       	 print, "updating fits header with trim values"
         ;
         ;    do the masking
         ;
         WCT_struct.trim_col = 0L   ; no columns trimmed
         ;WCT_struct.trim_row = 260L    ; trimmed first 260 rows
         WCT_struct.trim_row = yl		; trim first yl rows
         ;
         FXADDPAR, FITS_header, "TRIM-COL", WCT_struct.trim_col, "Mimir image trim - cols skipped from raw image edge"
         FXADDPAR, FITS_header, "TRIM-ROW", WCT_struct.trim_row, "Mimir image trim - rows skipped from raw image edge"
         ;
         result_image = image[xl:xu,yl:yu]
         size_result = SIZE(result_image)
         nx = size_result[1]
         ny = size_result[2]
         ;
         ;
         WCT_struct.ncol_result = nx
         WCT_struct.nrow_result = ny
         ;
         WCT_struct.result_image[0:nx-1, 0:ny-1] = result_image
         WCT_struct.result_image[nx:*, ny:*] = 0.0d0
         ;
         WCT_struct.result_flag = 1
         WCT_struct.raw_flag = 0
         ;
         WCT_struct.coadd = 0
         WCT_struct.file_mode = 0
         ;
         nlines = N_ELEMENTS(FITS_header)
         WCT_struct.result_FITS_lines = nlines
         WCT_struct.result_FITS_header[0:nlines-1] = FITS_header
         WCT_struct.result_FITS_header[nlines:*] = ''
         ;
         WCT_struct.result_file_root = file_name
        ;
        ;    plot
        ;
        WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
        S3_Display_On, Event
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        ;
        ;    look for continuum spectra if not Argon calibration
        ;
        if(WCT_struct.cal_mode ne 0) then begin
	        ;
	        WCT_print_text, Event, "Testing for Continuum Spectrum in Image"
	        ;
	        WCT_find_spectrum, result_image, spec_zone, sky_zone, spec_mode, IDATA=idebug
	        ;
	        ;    and mark plot if any found
	        ;
	        spec_count = TOTAL(spec_zone)
	        if(spec_count ne 0) then begin
	            window_no = WCT_struct.base_window_no
	            wset, window_no
	            ind = WHERE(spec_zone eq 1)
	            tvlct, [0, 255, 0, 0, 255], [0, 0, 255, 0, 255], [0, 0, 0, 255, 255], 10
	            for ispec = 0, spec_count-1 do begin
	               xx1= [0.,0.05]
	               xx2 = [0.95,1.0]
	               yy = [ind[ispec], ind[ispec]]
	               plots, xx1, yy/float(ny), /NORMAL, color=14
	               plots, xx2, yy/float(ny), /NORMAL, color=14
	               ;
	               ; and set those image rows to deleted data
	               ;
	               result_image[*,ind[ispec]] = WCT_struct.deleted_value
	            endfor
	        endif else begin
	        	print, "no trimming or spectrum search - (raw, reduced flags = ",$
	        			WCT_struct.raw_flag,WCT_struct.result_flag
	        endelse
	    endif
      endif
    endif else begin
       WCT_print_text, Event, "ERROR - no image loaded yet"
    endelse
    ;
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
end

;----------------------
pro S3_Coadd_On, Event
    ;
    ;  turn on hourglass
    ;
    WIDGET_CONTROL, /HOURGLASS
    ;
    ;  coadd list of images already read in
    ;
    ;   1. verify list is not empty
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    ;
    instrument = WCT_struct.instrument
    camera = WCT_struct.camera
    spec_mode = WCT_struct.spec_mode
    slit = WCT_struct.slit
    ;
    xl = WCT_struct.trim_col
    yl = WCT_struct.trim_row
    xu = WCT_struct.ncol_result + xl - 1
    yl = WCT_struct.nrow_result + yl - 1
    ;
    xcol = 1024
    xrow = 1026
    ;
    spec_mode = WCT_struct.spec_mode
    ;
    if(WCT_struct.file_mode le 1) then begin
       WCT_print_text, Event, "ERROR - No Images in the coadd list"
       ;print, "ERROR - No Images in the coadd list"
    endif else begin
       ;
       ; and that coadding *needs to be done*
       ;
       if(WCT_struct.coadd le 1) then begin
         WCT_print_text, Event, "ERROR - Coadding already done for this file list"
         ;print, "ERROR - Coadding already done for this file list"
       endif else begin
         ;
         ;    ok, list not empty, not already coadded, fetch it
         ;
         nfiles = WCT_struct.nfiles
         ;print, "Coadding ",nfiles," files"
         input_list = WCT_struct.files_multi[0:nfiles-1]
         ;
         ;       2. start coadder, turning on widget hourglass
         ;
         WCT_imcombine_no_bad_pix, Event, spec_mode, input_list, output_average_image, output_FITS_header, $
         							WCT_struct
         ;
         ;       3. leave resulting image in the "result" image area
         ;
         im_size = SIZE(output_average_image)
         ncol_out = im_size[1]
         nrow_out = im_size[2]
         ;
         WCT_struct.result_image[0:ncol_out-1,0:nrow_out-1] = output_average_image
         WCT_struct.result_image[ncol_out:*,nrow_out:*] = 0.0d0
         ;
         ;  if the coadded images already had TRIM-COL header keywords, grab them
         ;  and continue them into the WCT_structure
         ;
         trim_col = FXPAR(output_FITS_header, "TRIM-COL", COUNT=c_count)
         if(c_count ne 0) then WCT_struct.trim_col = trim_col
         trim_row = FXPAR(output_FITS_header, "TRIM-ROW", COUNT=r_count)
         if(r_count ne 0) then WCT_struct.trim_row = trim_row
         ;
         ;print, "after coadding, trim_col, trim_row = ", trim_col, trim_row
         ;
         n_FITS_lines = N_ELEMENTS(output_FITS_header)
         WCT_struct.result_FITS_lines = n_FITS_lines
         WCT_struct.result_FITS_header[0:n_FITS_lines-1] = output_FITS_header
         WCT_struct.result_FITS_header[n_FITS_lines:*] = ''
         ;
         WCT_struct.ncol_result = ncol_out
         WCT_struct.nrow_result = nrow_out
         ;
         ;    reset the coadd flag
         ;
         WCT_struct.coadd = 0
         ;
         ;  and the raw flag
         ;
         WCT_struct.raw_flag = 0
         WCT_struct.result_flag = 1
         ;
         ;  and plot
         ;
         WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
         S3_Display_On, Event
         WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
       endelse
    endelse
    ;
    ;  return the structure
    ;
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
end

;---------------------
pro S3_Save_On, Event
    ;
    ;  save the current image - build the name from the output directory
    ;   and the desired file name. But, if no output directory selected,
    ;   bring up the full path file chooser.
    ;
    WCT_print_text, Event, "Export Image Selected"
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    ;
    file_default = FILE_BASENAME(WCT_struct.result_file_root)
    if(WCT_struct.output_dir eq '') then begin   ; if user didn't select output directory
        path_out = WCT_struct.result_file_root
    endif else begin                    ; if they did
        path_out = WCT_struct.output_dir
    endelse
    file_out = dialog_pickfile(title="Export Image Name/Path",$
            PATH=path_out, FILE=file_default)
    ;
    ;   test for null string, indicating the user cancelled this operation
    ;
    if(file_out ne '') then begin
        ;
        ;   extract image from result image buffer
        ;
        ncol = WCT_struct.ncol_result
        nrow = WCT_struct.nrow_result
        ;
        image = WCT_struct.result_image[0:ncol-1,0:nrow-1]
        ;
        ;   and FITS header
        ;
        header = WCT_struct.result_FITS_header[0:WCT_struct.result_FITS_lines-1]
        ;
        FXADDPAR, header, "CAL-MODE", WCT_struct.cal_mode,"Mimir Calibration mode: 0 = Argon, 1 = OH night sky lines"
        FXADDPAR, header, "SPECMODE", WCT_struct.spec_mode,"Mimir Spectroscopy mode:0=JHK;1=K;2=J-3rd;3=J-4th;4=LM;5=HK;6=K17;7=IZYJ-1;8=IZYJ-2;9=BVR"
        FXADDPAR, header, "TRIM-ROW", WCT_struct.trim_row, "Mimir image trim - rows skipped from raw image edge"
        FXADDPAR, header, "TRIM-COL", WCT_struct.trim_col, "Mimir image trim - cols skipped from raw image edge"
       ;
;       print, "just before write and extracted from WCT_struct, trim_col, trim_row = ",WCT_struct.trim_col,$
;       			WCT_struct.trim_row
       	;
        if(WCT_struct.nfiles gt 0) then begin
            FXADDPAR, header, "COMMENT", "COADD image computed from the following elemental images:"
            FXADDPAR, header, "COMMENT", STRING(format='("   files directory = ",a)',FILE_DIRNAME(WCT_struct.files_multi[0]))
            for jfile = 0, WCT_struct.nfiles-1 do begin
               FXADDPAR, header, "COMMENT", STRING(format='(5x,"+",a)',FILE_BASENAME(WCT_struct.files_multi[jfile]))
            endfor
        endif
        ;
        ;   write file
        ;
        FXADDPAR, header, "FILENAME", FILE_BASENAME(file_out),"Filename"
        ;
        nFITS_lines = N_ELEMENTS(header)
        WCT_struct.result_FITS_lines = nFITS_lines
        WCT_struct.result_FITS_header[0:nFITS_lines-1] = header
        WCT_struct.result_FITS_header[nFITS_lines:*] = ''
        ;
        writefits, file_out, image, header
        ;
        WCT_print_text, Event, "Image Export Successful"
        ;
    endif else begin
       WCT_print_text, Event, "Operation Cancel Recognized - no file written"
    endelse
    ;
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
end

;-----------------------------------------------------------------
pro S4_launch_PU, Event, window_numbers, w_ID
    ;
    ;  launch the pop-up window and return the window numbers
    ;
    ;     create a heap variable structure and pointer to pass to the
    ;     pop up widget
    ;
    wStruct = {top_ID: -1L, middle_ID: -1L, bottom_ID: -1L, widget_ID: -1L}
    ;
    S4_ptr = PTR_NEW(wStruct)
    ;
    S4_struct_ptr = {ptr: S4_ptr}
    ;
    ;   call the pop up, passing the pointer
    ;
    WCT_S4_3_up_plot_pop_up, GROUP_LEADER=group_leader, _EXTRA = S4_struct_ptr
    ;
    ;print, "On return to S4_launch, S4 struct = ",*S4_ptr
    ;help, *S4_ptr, /struct
    ;
    ; put the window IDs and widget_ID into the main structure
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    ;
    ;   convert widget numbers to window numbers
    ;
    WIDGET_CONTROL, (*S4_ptr).top_ID, GET_VALUE=top_win_no
    WIDGET_CONTROL, (*S4_ptr).middle_ID, GET_VALUE=middle_win_no
    WIDGET_CONTROL, (*S4_ptr).bottom_ID, GET_VALUE=bottom_win_no
    ;
    WCT_struct.S4_top_win_no = top_win_no
    WCT_struct.S4_middle_win_no = middle_win_no
    WCT_struct.S4_bottom_win_no = bottom_win_no
    WCT_struct.S4_widget_ID = (*S4_ptr).widget_ID
    ;
    window_numbers = [top_win_no, middle_win_no, bottom_win_no]
    w_ID = (*S4_ptr).widget_ID
    ;
    ;   try to move the S4 pop up window to right hand side of display
    ;
    WIDGET_CONTROL, w_ID, XOFFSET=400     ; pixels
    ;
    ;   trial plots
    ;
    ;wset, WCT_struct.S4_top_win_no
    ;xx = indgen(100)
    ;yy = sin(xx)
    ;plot, xx, yy
    ;
    ;wset, WCT_struct.S4_middle_win_no
    ;yy = sin(xx/5.)
    ;plot, xx, yy
    ;
    ;wset, WCT_struct.S4_bottom_win_no
    ;yy = sin(xx/5)
    ;plot, xx, yy
    ;
    ;   resave the structure
    ;
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
end

;-----------------------------------------------------------------
pro S4_Find_Slit_Offset_On, Event
    WCT_print_text, Event, "Finding Slit Offset"
    ;
    ;  select the appropriate portion of the image to most easily
    ;  locate a bright Argon or OH line
    ;
    ;  compare gaussian fitted location to nominal value, report
    ;  difference
    ;
    ;   1. Unpack image
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    ;
    ;   get debug status from debug flag
    ;
    idata = WCT_struct.debug_flag     ; 1 for debug prints, 0 for not
    ;
    ncol = WCT_struct.ncol_result
    nrow = WCT_struct.nrow_result
    ;
    ;   test for raw only image
    ;
    if(ncol eq 0 or nrow eq 0) then begin
        print,"ERROR - ncol, nrow incorrect: ncol, nrow = ",ncol, nrow
        ncol = WCT_struct.ncol_raw
        nrow = WCT_struct.nrow_raw
        image = WCT_struct.raw_image[0:ncol-1, 0:nrow-1]
    endif else begin
        image = WCT_struct.result_image[0:ncol-1,0:nrow-1]
    end
    ;
    ;     and mode
    ;
    spec_mode = WCT_struct.spec_mode
    cal_mode = WCT_struct.cal_mode
    ;
    ;     and plot window number
    ;
    win_no = WCT_struct.base_window_no
    ;;
    ;       turn on hourglass
    ;
    WCT_print_text, Event, "Cleaning Image"
    ;
    WIDGET_CONTROL, /HOURGLASS
    ;
    ;   2. clean image, median filter it to enhance lines
    ;
    iskip = 0                ; turn off cleaning
    if(iskip eq 0) then begin
        clean_image = fix_bad_pixels(image)
        ;
        med_image = WCT_median_filter_image(clean_image,[21,0])
        ;
        dif_image = image - med_image
        ;
        ; chop off the negatives: turn off for LM
        ;
        if(WCT_struct.cal_mode ne 2) then begin
	        indices=where(dif_image lt -50.0,count)          ;
	        if(count gt 0) then dif_image[indices] = -50.0
	    endif
        ;
    endif else begin
       dif_image = image
    endelse
    ;
    ;   save the differenced image in the structure for later use
    ;
    dif_size = SIZE(dif_image)
    ncol_dif = dif_size[1]
    nrow_dif = dif_size[2]
    WCT_struct.ncol_dif = ncol_dif
    WCT_struct.nrow_dif = nrow_dif
    WCT_struct.dif_image[0:ncol_dif - 1, 0:nrow_dif - 1] = dif_image
    WCT_struct.dif_image[ncol_dif:*, nrow_dif:*] = 0.0d0
    ;
    spec = fltarr(ncol)                   ; build the MFM spectrum
    ;
    ;	if fat-eared, only look at top and bottom
    ;
    if(WCT_struct.instrument eq 0 and WCT_struct.slit eq 4) then begin
    	y0 = 1
    	y1 = 40
    	y2 = nrow_dif-40
    	y3 = y2 + 39
    	xoff = 3				; shift from bottom to top
    endif else begin
    	nd2 = nrow_dif / 2
    	y0 = 0
    	y1 = nd2-1
    	y2 = nd2
    	y3 = nrow_dif - 1
    	xoff = 0
    endelse
    ;
    for ic = 0, ncol-1-xoff do begin
        v1 = REFORM(dif_image[ic,y0:y1])
        v2 = REFORM(dif_image[(0 > (ic+xoff)) < (ncol_dif-1),y2:y3])
        vector = [v1,v2]
        zz = median_filtered_mean(vector)
        spec[ic] = zz[0]
    endfor
    ;
    zz = median_filtered_mean(spec)
    ind = WHERE(spec lt zz[0]-3.0*zz[1],icount)
    if(icount gt 0 and WCT_struct.cal_mode ne 2) then begin
    	spec[ind] = zz[0]
    endif
    ;
;    window, 22
;    plot, spec
    ;
    case spec_mode of
        0: disp_in = WCT_struct.dispersion_library.jhk_jhk_dispersion       ; 1st order JHK-Spec
        1: disp_in = WCT_struct.dispersion_library.jhk_k_dispersion        ; 1st order K-spec
        2: disp_in = WCT_struct.dispersion_library.lm_3rd_dispersion        ; 3rd order J-spec
        3: disp_in = WCT_struct.dispersion_library.lm_4th_dispersion        ; 4th order J-spec
        4: disp_in = WCT_struct.dispersion_library.lm_1st_dispersion        ; LM-spec
        5: disp_in = WCT_struct.dispersion_library.jhk_hk_dispersion		; HK-Spec
        6: disp_in = WCT_struct.dispersion_library.lm_2nd_k17_dispersion		; K17-Spec
        7: disp_in = WCT_struct.dispersion_library.jhk_izyj_1st_dispersion	; IZYJ 1st order
        8: disp_in = WCT_struct.dispersion_library.jhk_izyj_2nd_dispersion	; IZYJ 2nd order
        9: disp_in = WCT_struct.dispersion_library.jhk_h17_dispersion		; H17-Spec
        10: disp_in = WCT_struct.dispersion_library.jhk_iz17_2nd_dispersion	; IZ17-Spec
        10: disp_in = WCT_struct.dispersion_library.bvr_vph_dispersion       ; BVR/VPH spec
    endcase
    ;
    ;   check for this image already calibrated to linear dispersion
    ;
    ;     first check the RESULT buffer FITS file, then the RAW buffer FITS file
    ;
    if(WCT_struct.result_FITS_lines ne 0) then begin
         header = WCT_struct.result_FITS_header[0:WCT_struct.result_FITS_lines-1]
         ;
         CRVAL1 = FXPAR(header, "CRVAL1", COUNT=ii)
         CD1_1 = FXPAR(header, "CD1_1", COUNT=jj)
         if(ii+jj eq 2) then begin
              disp_in = [double(CRVAL1), double(CD1_1), 0.0d0]
              if(idata ne 0) then print, "Found calibrated data in RESULT image buffer"
         endif
    endif else begin
        ;
        if(WCT_struct.raw_FITS_lines ne 0) then begin
            header = WCT_struct.raw_FITS_header[0:WCT_struct.raw_FITS_lines-1]
            ;
            CRVAL1 = FXPAR(header, "CRVAL1", COUNT=ii)
            CD1_1 = FXPAR(header, "CD1_1", COUNT=jj)
            if(ii+jj eq 2) then begin
               disp_in = [double(CRVAL1), double(CD1_1), 0.0d0]
               if(idata ne 0) then print, "Found calibrated data in RAW image buffer"
            endif
        endif
    endelse
    ;
    if(idata ne 0) then print,"Disp_in = ",disp_in[0], disp_in[1], disp_in[2]
    ;
    ;  locate the brightest line in a range of columns for the selected mode
    ;     finding the brightest line isn't necessarily the best way to
    ;     find the slit offset, but it is expedient...
    ;
     case cal_mode of
        0: begin
              ;
                ; Argon
                ;
              case spec_mode of
               0: brightest = 1.69452d0     ; JHK brightest line
			   1: brightest = 2.0621860d0	; K-Spec
               2: brightest = 1.2956659d0	; J 3rd
               3: brightest = 1.2956659d0	; J 4th
               4:							; LM
               5: brightest = 1.69452d0		; HK
               6: brightest = 2.062186d0	; K17 spec
               7: brightest = 1.2956659d0	; IZYJ 1st
               8: brightest = 0.9657786d0	; IZYJ 2nd
               9: brightest = 1.6940580d0	; H17-Spec
               10: brightest = 0.9122967d0  ; IZ17
               11:
              endcase
          end
          ;
        1: begin
            ;
               ; OH
               ;
               case spec_mode of
                      0: brightest = 1.58366d0
                      1: brightest = 2.00110d0
                      2: brightest = 1.29087d0   ; not very good line
                      3: brightest = 1.29087d0   ; not very good line
                      4:
                      5: brightest = 1.58366d0
                      6:
                      7:
                      8:
                      9:
                      10:
                      11:
               endcase
           end
        2: begin								; LM sky case
        	case spec_mode of
        		4: brightest = 3.31594d0
        	endcase
           end
    endcase
    ;
    ;  find nominal line column for zero slit center offset
    ;
    if(disp_in[2] ne 0.0d0) then begin
        col_cen = (-disp_in[1]/(2.0d0*disp_in[2])) + (1.0d0/(2.0d0*disp_in[2]))* $
               sqrt(disp_in[1]^2 - (4.0d0*disp_in[2]*(disp_in[0]-brightest)))
    endif else begin
       col_cen = 0.5d0 + (brightest - CRVAL1) / CD1_1
       if(col_cen eq 0) then print, "ERROR - col_cen = 0 in S4 Find Slit Offset"
    endelse
    ;
    if(idata ne 0) then print, "for brightest line =",brightest," microns, the nominal center channel = ",col_cen
    ;
    ;
    ;  do passes: on first pass go +/- 15 pixels, on second go +/- 8
    ;	for fat-eared slit, start at +/- 50
    ;
    icol_cen = ROUND(col_cen)
    fit_col_cen = icol_cen
    ;
    for ipass = 0, 1 do begin
        ;
        if(ipass eq 0) then begin
        	if(WCT_struct.instrument eq 0 and WCT_struct.slit eq 4) then begin
        		offset = 50
        	endif else begin
        		offset = 20			; was 15 - swapped to 25 for Noel R. - DPC 20090930
        	endelse
        endif else begin
        	offset = 8
        endelse
        ;        ;
        bright_spec = spec[fit_col_cen-offset:fit_col_cen+offset]
        x_bright = findgen(2*offset+1) + float(fit_col_cen) - offset + 0.5
        max_b = max(bright_spec)
        col_b_ind = WHERE(bright_spec[*] eq max_b)
        col_b = x_bright[col_b_ind]
        wid_b = 1.0
        offset = 0.0
        if(ipass eq 0) then begin
           est = [max_b, col_b, wid_b, offset]
        endif else begin
           est = a
        endelse
        ;
        yfit = wGAUSSFIT(x_bright,bright_spec,a,NTERMS=4,SIGMA=sigma_a,ESTIMATES=est)
        ;
        fit_col_cen = ROUND(a[1])
        if(idata ne 0) then print, "ipass=",ipass," a(fit) = ",a[0],a[1],a[2],a[3]
    endfor
    ;
    wset, win_no
    ;
    ; reset plot window size
    ;
    base_plot_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="base_plot")
    if(idata ne 0) then begin
        base_geo = WIDGET_INFO( base_plot_id, /GEOMETRY)
        help, base_geo, /struct
    endif
    WIDGET_CONTROL, base_plot_id, DRAW_XSIZE=WCT_struct.base_xsize, DRAW_YSIZE=290, SCR_YSIZE=308
    ;
    if(idata gt 0) then print, "resetting base plot size"
    ;
    erase
    ;
    xmin = min(x_bright)
    xmax = max(x_bright)
    dx = xmax - xmin
    ymax = a[0]+a[3]
    ymin = a[3]
    dy = ymax - ymin
    ymax = ymax+0.2*dy
    ymin = ymin-0.1*dy
    dy = ymax-ymin
    ;
    device, decomposed=0
    black = 14
    loadct,0, /SILENT
    tvlct, [0, 255, 0, 0, 255], [0, 0, 255, 0, 255], [0, 0, 0, 255, 255], 10
   ;
    plot, x_bright, bright_spec, xrange=[xmin, xmax], yrange=[ymin, ymax], psym=4, xstyle=1, $
        ystyle=1, /NODATA
    oplot, x_bright, yfit, color=13                ; fit (blue)
    oplot, x_bright, bright_spec, color=11, psym=4  ; spectrum (red) - points only
    xc = [col_cen, col_cen]
    yc = [ymin, ymin+0.3*dy]
    oplot, xc, yc, thick=4, color=12               ; nominal center
    ;
    ;   compute correction to disp_in[0]
    ;
    slit_offset = a[1] - col_cen
    xyouts, fit_col_cen, ymax-0.1*dy, STRING(format='("Slit Offset = ",f6.2," +/-",f5.2," pixels")',$
                            slit_offset,sigma_a[1]), charsize=2.0
    ;
    ;   add mode to plot
    ;
    xc = xmin + 0.8*dx
    yc = ymax - 0.4*dy
    case cal_mode of
    	0: cal = "Argon Mode"
    	1: cal = "OH Mode"
    	2: cal = "LM Mode"
    endcase
    xyouts, xc, yc, cal, color=12, charsize=2.0, charthick=3.0
    ;
    ;  put into text window
    ;
    text_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S4_Find_Slit_text_box")
    WIDGET_CONTROL, text_id, SET_VALUE=STRING(format='(f6.2)',slit_offset)
    ;
    ;   compute the correction factor for cal columns vs science columns - DPC 20070803
    ;
    ;col_shift = 4.0 - slit_offset                   ; this must be added to the column number of each argon -or -
    ;zero_lambda_shift = col_shift * disp_in[1]     ; to be subtracted from the final dispersion equations before applying to science images
    ;
    ;   compute additive change for disp_in[0] term
    ;
    disp_in[0] = disp_in[0] - slit_offset * disp_in[1]
    ddsp = -1.0 * slit_offset * disp_in[1]
    ;
    WCT_print_text, Event, STRING(format='("Correction to zero column wavelength = ",f12.7," microns")',ddsp)
    ;
    ;print, "slit offset = ",zero_lambda_shift," subtracted from science images"
    ;
    ;   save corrected nominal dispersion equation and exit
    ;
    WCT_struct.disp_0 = disp_in[0]
    WCT_struct.disp_1 = disp_in[1]
    WCT_struct.disp_2 = disp_in[2]
    ;
    ;   and put into text boxes on GUI
    ;
   disp_0_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S4_disp_0_text_box")
   WIDGET_CONTROL, disp_0_id, SET_VALUE=STRING(format='(f9.6)',disp_in[0])
   disp_1_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S4_disp_1_text_box")
   WIDGET_CONTROL, disp_1_id, SET_VALUE=STRING(format='(e13.5)',disp_in[1])
   disp_2_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S4_disp_2_text_box")
   WIDGET_CONTROL, disp_2_id, SET_VALUE=STRING(format='(e13.5)',disp_in[2])
   ;
   ;       save the spectrum in the main structure
   ;
   nch = N_ELEMENTS(spec)
   x_spec = dindgen(nch)+0.5d0
   WCT_struct.nch = nch
   WCT_struct.x_spec[0:nch-1] = x_spec
   WCT_struct.y_spec[0:nch-1] = spec
   ;
    ;
    ;  repack structure and end
    ;
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
end

;-----------------------------------------------------------------
pro S4_Refine_On, Event
    ;
    WCT_print_text, Event, "Refining Overall Dispersion Equation"
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    ;
    ; debug
    ;
    idata = WCT_struct.debug_flag     ; 1 for lots of debug; 0 for none
    ;
    ;   get spectrum
    ;
    nch = WCT_struct.nch
    if(nch gt 0) then begin
        x_spec = WCT_struct.x_spec[0:nch-1]
        y_spec = WCT_struct.y_spec[0:nch-1]
        ;
        ;     and mode
        ;
        spec_mode = WCT_struct.spec_mode
        ;
        wave_limits = WCT_struct.wave_limits

        ;
        this_wave_limits = REFORM(wave_limits[spec_mode,*])
            ;
            ;
            cal_mode = WCT_struct.cal_mode
            ;
            ;  load lines and line quality vectors
            ;
            if(cal_mode eq 0) then begin
                 ;
                 ;    argon
                 ;
                 lines = WCT_struct.argon.wave
                 line_qual = REFORM(WCT_struct.argon.qual[spec_mode,*])
                 nlines = N_ELEMENTS(lines)
                 ;
                 WCT_print_text, Event, STRING(format='("Loaded Argon Line List of ",i," lines")',nlines)
            endif else begin
            	if(cal_mode eq 1) then begin
	                 ;
	                 ;    OH
	                 ;
	                 lines = WCT_struct.OH.wave
	                 line_qual = REFORM(WCT_struct.OH.qual[spec_mode,*])
	                 nlines = N_ELEMENTS(lines)
	                 WCT_print_text, Event, STRING(format='("Loaded OH Line List of ",i," lines")',nlines)
                endif else begin
                 	;
                 	;	LM
                 	;
                 	lines = WCT_struct.LM.wave
	                line_qual = REFORM(WCT_struct.LM.qual[spec_mode,*])
	                nlines = N_ELEMENTS(lines)
	                WCT_print_text, Event, STRING(format='("Loaded LM Line List of ",i," lines")',nlines)
	            endelse
           endelse
           ;
           ; load the in-going dispersion equation vector
           ;
           disp_in = [WCT_struct.disp_0, WCT_struct.disp_1, WCT_struct.disp_2]
           ;
            ;   call dispersion refiner
            ;
            WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
            ;
            case cal_mode of
            	0: fit_width = 2.5     ; limit to +/- 2.5 ch
            	1: fit_width = 2.5
            	2: fit_width = 5.0
            endcase
            ;
            WCT_refine_dispersion, Event, disp_in, disp_out, wave_fit, x_spec, y_spec, cal_mode, lines, $
                   line_qual, this_wave_limits, IDATA=idata, FIT_WIDTH=fit_width, rms, NOPLOT=noplot, good_fit
            ;
            WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
            ;
            ;  save the fitted dispersion
            ;
            WCT_struct.disp_0 = disp_out[0]
            WCT_struct.disp_1 = disp_out[1]
            WCT_struct.disp_2 = disp_out[2]
            ;
            ;  update the dispersion text windows
            ;
            disp_0_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S4_disp_0_text_box")
            WIDGET_CONTROL, disp_0_id, SET_VALUE=STRING(format='(f9.6)',disp_out[0])
            disp_1_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S4_disp_1_text_box")
            WIDGET_CONTROL, disp_1_id, SET_VALUE=STRING(format='(e13.5)',disp_out[1])
            disp_2_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S4_disp_2_text_box")
            WIDGET_CONTROL, disp_2_id, SET_VALUE=STRING(format='(e13.5)',disp_out[2])
            ;
    endif else begin
       WCT_print_text, Event, "No 1-D spectrum was found. Run slit offset finder first"
    endelse
    ;
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
end

;-----------------------------------------------------------------
pro S4_Done, Event
    WCT_print_text, Event, "Closing Tab 4, Opening Tab 5"
    ;
    tab_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="Step_tab")
    WIDGET_CONTROL, tab_ID, SET_TAB_CURRENT=5
    New_Tab, Event
end

;-------------------------- S5 Compute Slit Shapes vs Row ---------
pro S5_Zone_size_text_box_input, Event
    ;
    ;  text has been entered into the S5 Zone size text box
    ;
    ;  grab it, interpret it, but don't change the zone size
    ;  until the button is pressed.
    ;
    text_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_Zone_size_text_box")
    ;
    WIDGET_CONTROL, text_box_id, GET_VALUE=text_string
    ;
    ;print, "S5 Zone size text box string = ",text_string
    ;
    ;  trim and interpret
    ;
    text_string = STRTRIM(text_string,2)
    reads, text_string, new_zone_size
    ;
    ;print, "new zone size = ",new_zone_size
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    ;
    if(WCT_struct.instrument eq 0 and WCT_struct.slit eq 4) then begin
    	WCT_struct.new_zone_size = 2
    endif else begin
    	WCT_struct.new_zone_size = new_zone_size
    endelse
    ;
    zone_size = WCT_struct.zone_size
    nrow = WCT_struct.nrow_dif
    ;
    ;print, "In S5 Change Zone Size, zone_size, new_zone_size = ",zone_size, new_zone_size
    ;
    if(new_zone_size gt 0 and new_zone_size ne zone_size and new_zone_size le nrow) then begin
        WCT_struct.zone_size = new_zone_size
        WCT_print_text, Event, STRING(format='("Changing zone size from ",i3," to ",i3)',zone_size, new_zone_size)
    endif
    ;
    if(nrow gt 0) then begin
       ;
       ; how many zones?
       ;
       if(WCT_struct.instrument eq 0 and WCT_struct.slit eq 4) then begin
       	nzones = 2
       endif else begin
       	nzones = CEIL(nrow/zone_size)
       endelse
       ;
       text_box_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_Zone_text_box")
       WIDGET_CONTROL, text_box_ID, SET_VALUE=STRING(format='(i4)',nzones)
       ;
       ; save in structure
       ;
       WCT_struct.nzones = nzones
       ;
       ; load the first, last row per zone
       ;
       if(WCT_struct.instrument eq 0 and WCT_struct.slit eq 4) then begin
       	WCT_struct.row_zones[0,0] = 1
       	WCT_struct.row_zones[0,1] = 38
       	WCT_struct.row_zones[1,0] = WCT_struct.nrow_result - 39
       	WCT_struct.row_zones[1,1] = WCT_struct.nrow_result - 2
       endif else begin
       	for izone = 0, nzones-1 do begin
         zstart = (izone * zone_size) > 0
         zstop = (zstart + zone_size) < (nrow-1)
         ;
         WCT_struct.row_zones[izone,0] = zstart
         WCT_struct.row_zones[izone,1] = zstop
        endfor
       endelse
       ;
       ; clear remainder
       ;
       WCT_struct.row_zones[nzones:*,*] = -1
       ;
       ;    clear LED-like stack
       ;
       led_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_Quality_plot_box")
       ;
       led_colors = strarr(nzones)
       for izone = 0, nzones-1 do begin
        led_colors[izone] = "gray"
       endfor
       ;
       WCT_LED_stack, led_id, led_colors
       ;
    endif
    ;
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    WCT_print_text, Event, "Zone Size, Zone Numbers Changed"
end

;--------
pro S5_Start_Fit_On, Event
    ;
    ;  recover the data structure
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    ;
    ;   debug
    ;
    idata = WCT_struct.debug_flag
    ;
    deleted_data = -1.0e6
    ;
    ;   get the LED bar graph widget id
    ;
    led_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_Quality_plot_box")
    ;
    ;   get image size
    ;
    nrow = WCT_struct.nrow_dif
    ncol = WCT_struct.ncol_dif
    ;
    ;   and the image
    ;
    image = WCT_struct.dif_image[0:ncol-1, 0:nrow-1]
    ;
    ;  get number of lines and number of zones
    ;
    ntotalzones = WCT_struct.nzones
    nzones = ntotalzones
    ;
    ;   spec mode
    ;
    spec_mode = WCT_struct.spec_mode
    ;
    ;   clear LED stack again
    ;
    led_colors = strarr(nzones)
    for izone = 0, nzones-1 do begin
      led_colors[izone] = "gray"
    endfor
    ;
    WCT_LED_stack, led_id, led_colors
    ;
    row_zones = WCT_struct.row_zones
    ;
    cal_mode = WCT_struct.cal_mode
    case cal_mode of
    	0: begin
	        wave = WCT_struct.Argon.wave
	        qual = REFORM(WCT_struct.Argon.qual[spec_mode,*])
	       end
	    1: begin
	        wave = WCT_struct.OH.wave
        	qual = REFORM(WCT_struct.OH.qual[spec_mode,*])
           end
        2: begin
        	wave = WCT_struct.LM.wave
        	qual = REFORM(WCT_struct.LM.qual[spec_mode,*])
           end
    endcase
    ;
    ;   load the fit limits by spec_mode
    ;
    this_wave_limits = REFORM(WCT_struct.wave_limits[spec_mode,*])
    ;
    line_ind = WHERE(qual eq 1 and abs(wave) ge this_wave_limits[0] and abs(wave) le this_wave_limits[1], nlines)
    WCT_print_text, Event, STRING(format='("Fitting to ",i," Calibration lines")',nlines)
    wave_line = wave[line_ind]
    ;
    ;   clear the line fit LED bar graph
    ;
    led_line_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_Slit_Fit_Quality_plot_box")
    led_line_colors = strarr(nlines)
    for il = 0, nlines-1 do begin
        led_line_colors[il] = "red"
    endfor
    WCT_LED_stack, led_line_id, led_line_colors
    ;
    row_fit = dblarr(3,nlines)                   ; to hold the final fit
    row_unc_fit = dblarr(3,nlines)             ; to hold uncertainties in fit   DPC  4/12/07
    row_data = dblarr(2,ntotalzones,nlines)        ; to hold the zonal data [0=fit,1=unc]
    row_value = fltarr(ntotalzones)
    ;
    disp_in = dblarr(3)
    ;
    ;   LED bar graph colors
    ;
    led_colors = strarr(nzones)
    for iz = 0, nzones-1 do begin
        led_colors[iz] = "red"
    endfor
    WCT_LED_stack, led_id, led_colors
    ;
    ;   set up channel spectrum
    ;
    x_spec = findgen(ncol) + 0.5
    ;
    for izone = 0, nzones-1 do begin
        led_colors[izone] = "orange"
        WCT_LED_stack, led_id, led_colors
        ;
        ir0 = row_zones[izone,0]
        ir1 = row_zones[izone,1]
        row_value[izone] = float(ir0 + ir1) / 2.0d0     ; save effective row #
        ;
        if(idata eq 1) then print, "Fitting Night Sky Lines for rows ",ir0," to ",ir1
        ;
        ;    compress to a spectrum
        ;
        spec = dblarr(ncol)
        for ic = 0, ncol-1 do begin
           vec = REFORM(image[ic:ic,ir0:ir1])
           zz = median_filtered_mean(vec)
           spec[ic] = zz[0]
        endfor
        ;
        led_colors[izone] = "yellow"
        WCT_LED_stack, led_id, led_colors
        ;
        ;    improve the fit
        ;
        ;       first, reload the nominal dispersion equation
        ;
        disp_in[0] = WCT_struct.disp_0
        disp_in[1] = WCT_struct.disp_1
        disp_in[2] = WCT_struct.disp_2
        ;
        disp_in_save = disp_in
        ;
        zone_fit = dblarr(1,1)
        WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
        ;
        case cal_mode of
        	0: fit_width=3.5
        	1: fit_width=3.5
        	2: fit_width=6.5
        endcase
        ;
        ;   for +/- fit_width channel offsets, find the lowest line fit dispersion
        ;
        dch = CEIL(fit_width)+1
        rms_min = 1.0e6
        m_dch = 0.0
        noplot = 1
        ;
        ntries = 2*dch + 1
        ;
        results = fltarr(ntries,2)		; 0 = rms, 1 = nlines
        ;
        for idch = -dch, dch do begin
            ;
         disp_in = disp_in_save
         disp_in[0] = disp_in[0] + float(idch) * disp_in[1]
            ;
            WCT_refine_dispersion, Event, disp_in, disp_out, wave_fit, x_spec, spec, $
                         cal_mode, wave, qual, this_wave_limits, SAVE_ARRAY=zone_fit, $
                         FIT_WIDTH=fit_width, IDATA=idata, rms, NOPLOT=noplot, good_fit
            ;
            ;print, "on idch = ",idch," rms = ",rms
            ;
            n_lines_fit = good_fit
            ;
            results[idch+dch,0] = rms
            results[idch+dch,1] = n_lines_fit
            ;
            ;print, "izone ",izone," idch= ",idch," rms = ",rms," n_lines_fit = ",n_lines_fit
            ;
        endfor
        ;
        ;	select best fit from those with n_lines within 1 of maximum, and of those,
        ;	the one with the lowest rms
        ;
        n_lines_max = 0
        ;
        for it = 0, ntries-1 do begin
        	n_lines_max = n_lines_max > results[it,1]
        endfor
        ;
        ;print, "n_lines_max = ",n_lines_max
        ;
        if(n_lines_max ne 0) then begin
        	;
        	;	select group to consider
        	;
        	ind = WHERE(results[*,1] ge (n_lines_max - 1),count)
        	;
        	;print, "Found ",count," cases with n_lines within 1 of max"
        	;
        	if(count ge 1) then begin
        		;
        		;	select smallest rms of these
        		;
        		rms_min = 1.0e6
        		for ic = 0, count-1 do begin
        			if(results[ind[ic],0] lt rms_min) then begin
        				rms_min = results[ind[ic],0]
        				m_dch = ind[ic] - dch
        			endif
        		endfor
        		;
            endif
            ;
            ;print, "lowest rms = ",rms_min," at m_dch = ",m_dch
            ;
        endif
        ;
        if(idata gt 0) then print, "best rms = ",rms_min," at ",m_dch," channel offset"
        ;
        ;   rerun with best offset
        ;
        disp_in = disp_in_save
        disp_in[0] = disp_in[0] + float(m_dch) * disp_in[1]
        ;
        WCT_refine_dispersion, Event, disp_in, disp_out, wave_fit, x_spec, spec, $
                         cal_mode, wave, qual, this_wave_limits, SAVE_ARRAY=zone_fit, $
                         FIT_WIDTH=fit_width, IDATA=idata, rms, NOPLOT=noplot, good_fit
        save_zone_fit = zone_fit
        ;
        ;
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        ;
        led_colors[izone] = "green"
        WCT_LED_stack, led_id, led_colors
        ;
        ;    most interested in the actual column fits, not the dispersion equation
        ;
        row_data[0,izone,*] = zone_fit[1,*]     ; column fits
        row_data[1,izone,*] = zone_fit[2,*]     ; column fit uncer
        wave_line[*]        = zone_fit[0,*]             ; line wavelengths
       ;
    endfor
    ;
    ;row_data = row_data[*,0:nzones-1,*]
    ;row_value = row_value[0:nzones-1]
    ;
    ;   All the different row zones are done. Find Row dependence of each spectral line
    ;
    ;   now, for each fitted spectral line, clean the data of interlopers
    ;   then fit the row behavior
    ;
    ;for iline = 0, nlines-1 do begin
    ;
    line_good_flag = intarr(nlines)
    ;
    for iline = 0, nlines-1 do begin
       line_good_flag[iline] = 1		; starting value
       ;
       ; update the line text box
       ;
       box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_Line_text_box")
       WIDGET_CONTROL, box_id, SET_VALUE=STRING(format='(i2)',iline+1)
       ;
       ;    and the bar graph
       ;
       led_line_colors[iline] = "yellow"
       WCT_LED_stack, led_line_id, led_line_colors
       ;
       ;  collect the data into a vector
       ;
       fit_col = REFORM(row_data[0,*,iline])          ; collect all row zones values for line's column
       fit_unc = REFORM(row_data[1,*,iline])
       ;
       ;    test, remove failed fits
       ;
       ;print, "for line number = ",iline," wavelength = ",wave_line[iline]
       ;print, "Fit_col = ",fit_col
       ;print, "Fit_unc = ",fit_unc
       ;
       ind_fit = WHERE(fit_unc lt 5.0,fit_count)
       if(WCT_struct.instrument eq 0 and WCT_struct.slit eq 4) then begin
       	if(fit_count ge 2) then begin
       		count_good = 2
       		fit_col = fit_col[0:1]
       		fit_unc = fit_unc[0:1]
       		ind_good = [0, 1]
       	endif else begin
       		fit_col[*] = 500.
       		fit_unc[*] = 1000.0
       		count_good = 0
       		line_good_flag[iline] = 0
       	endelse
       endif else begin
       	   if(fit_count gt 4) then begin
	            ;print, "fit count = ",fit_count
	            fit_col = fit_col[ind_fit]
	            fit_unc = fit_unc[ind_fit]
	            ;**
		        ;print, "Good Fit_col = ",fit_col
		        ;print, "Good Fit_unc = ",fit_unc
		        ;
		        zz = median_filtered_mean(fit_col)
		        ;print, "mean, sigma = ",zz
		        zz[1] = zz[1] > 0.5                   ; trap dispersion at no less than 1/2 of a pixel
		        ;
		        ;  test for crazy fits, only fit good ones
		        ;
		        ind_good = WHERE(abs(fit_col-zz[0])/zz[1] le 6.,count_good)
		        ;print, "count_good = ",count_good
	       endif else begin
	       		;
	       		;	but, if fewer than 5 zones fit the line, kill the line
	       		;
	       		fit_col[*] = 500.
	       		fit_unc[*] = 1000.0
	       		count_good = 0
	       		line_good_flag[iline] = 0		;turn off this line
	       endelse
	   endelse
       ;
       if(WCT_struct.instrument eq 0 and WCT_struct.slit eq 4) then begin
       	;
       	if(count_good eq 2) then begin
       		a = LINFIT(row_value[ind_good],fit_col[ind_good],MEASURE_ERRORS=fit_unc[ind_good],$
                   SIGMA=s_a, /DOUBLE)
       		;
       		a = [a,0.0]
       		s_a = [s_a, 0.0]
       		;
       		if(a[0] ge 0. and a[0] le ncol-1 and a[1] gt -0.05 and a[1] lt 0.05 and a[2] gt -0.005 and a[2] lt 0.005) then begin
                row_fit[0,iline] = a[0]
                row_fit[1,iline] = a[1]
                row_fit[2,iline] = a[2]
                row_unc_fit[0,iline] = s_a[0]
                row_unc_fit[1,iline] = s_a[1]
                row_unc_fit[2,iline] = s_a[2]
                ;
                led_line_colors[iline] = "green"
                WCT_LED_stack, led_line_id, led_line_colors
            endif else begin
                ;
                led_line_colors[iline] = "red"
                WCT_LED_stack, led_line_id, led_line_colors
                ;
                ;print, "Bad fit for line ",wave_line[iline]," a[0-2] = ",a
                row_fit[0,iline] = 1.0
                row_fit[1,iline] = 0.014
                row_fit[2,iline] = 0.0
                row_unc_fit[0,iline] = 10.
                row_unc_fit[1,iline] = 1.
                row_unc_fit[2,iline] = 1.
                ;
                line_good_flag[iline] = 0
            endelse
		endif else begin
			WCT_print_text, Event, STRING(format='("ERROR - No good line fits for iline = ",i2)',iline)
           led_line_colors[iline] = "red"
            WCT_LED_stack, led_line_id, led_line_colors
            ;
            row_fit[*,iline] = -99.0
            row_unc_fit[*,iline] = 1.0e6
            ;
            line_good_flag[iline] = 0
        endelse
	   endif else begin
		;
	       if(count_good gt 3) then begin
	            ;
	            ;  trap uncertainties of fitted columns to no smaller than 0.333 pixels to allow all row zones to
	            ;  influence the fit, not just the strongest, central ones
	            ;
	            fit_unc = fit_unc > 0.333
	           ;
	           ;   fit line center versus row number
	           ;
	           a = POLY_FIT(row_value[ind_good],fit_col[ind_good],2,MEASURE_ERRORS=fit_unc[ind_good],$
	                   SIGMA=s_a, /DOUBLE, STATUS=pstatus)
	           ;
	           ;   test for bad fit: wild offset, large slope, huge curvature
	           ;
	           if(a[0] ge 0. and a[0] le ncol-1 and a[1] gt -0.05 and a[1] lt 0.05 and a[2] gt -0.005 and a[2] lt 0.005) then begin
	                row_fit[0,iline] = a[0]
	                row_fit[1,iline] = a[1]
	                row_fit[2,iline] = a[2]
	                row_unc_fit[0,iline] = s_a[0]
	                row_unc_fit[1,iline] = s_a[1]
	                row_unc_fit[2,iline] = s_a[2]
	                ;
	                led_line_colors[iline] = "green"
	                WCT_LED_stack, led_line_id, led_line_colors
	                ;
	                ;print, "Good fit for line ",wave_line[iline]," a[0]= ",a[0],"+/-",s_a[0]," a[1]= ",a[1],"+/-",s_a[1]," a[2]= ",a[2],"+/-",s_a[2]
	            endif else begin
	                ;
	                led_line_colors[iline] = "red"
	                WCT_LED_stack, led_line_id, led_line_colors
	                ;
	                ;print, "Bad fit for line ",wave_line[iline]," a[0-2] = ",a
	                row_fit[0,iline] = 1.0
	                row_fit[1,iline] = 0.014
	                row_fit[2,iline] = 0.0
	                row_unc_fit[0,iline] = 10.
	                row_unc_fit[1,iline] = 1.
	                row_unc_fit[2,iline] = 1.
	                ;
	                line_good_flag[iline] = 0
	            endelse
	        endif else begin
	           WCT_print_text, Event, STRING(format='("ERROR - No good line fits for iline = ",i2)',iline)
	           led_line_colors[iline] = "red"
	            WCT_LED_stack, led_line_id, led_line_colors
	            ;
	            row_fit[*,iline] = -99.0
	            row_unc_fit[*,iline] = 1.0e6
	            ;
	            line_good_flag[iline] = 0
	        endelse
	        ;
	   endelse
    endfor
    ;
    ;	number of surviving good lines?
    ;
    ind_good_lines = WHERE(line_good_flag eq 1, n_good_lines)
    ;
    ;print," number of good lines = ",n_good_lines," out of ",nlines," tested"
    ;
    case cal_mode of
    	0: min_lines = 8
    	1: min_lines = 8
    	2: min_lines = 5
    endcase
    ;
    if(n_good_lines lt min_lines) then begin
    	print,"Fewer than ",min_lines," good spectral lines to fit -- warning"
;    	stop
    endif
    ;
    ;   Create model spectral lines and plot their curvatures into the Slit plot window
    ;
    ;   Then, compute mean fit and residuals for each line -- flag lines with high residuals
    ;
    c_model = dblarr(nzones, nlines)
    ;
    ;     make the reference row the center of all the zones
    ;
    row_ref = (row_value[0] + row_value[nzones-1])/2.0d0
    ;
    for iline = 0, nlines-1 do begin
    	;
        ; don't plot real stinkers
        ;
    	if(line_good_flag[iline] eq 1) then begin
        	for izone = 0, nzones-1 do begin
	           irow = row_value[izone]
	           c_model[izone, iline] = row_fit[1,iline]*(irow - row_ref) + $
	                     row_fit[2,iline]*(irow*irow - row_ref*row_ref); remove offset term
	       endfor
        endif else begin
        	c_model[*,iline] = -99.0
        endelse
    endfor
    ;
    ;     compute mean model
    ;
    ave_slit = fltarr(nzones)
    ave_slit_sig = fltarr(nzones)
    ;
    for izone = 0, nzones-1 do begin
    	c_vec = REFORM(c_model[izone,*])
    	ind_good = WHERE(c_vec[*] ne -99.0, mcount)
    	if(mcount gt 0) then begin
	        vec = c_vec[ind_good]
	        ;
	        ;   test for all zeros
	        ;
	        ind = WHERE(vec ne 0, count)
	        if(count gt 0) then begin
	            zz = median_filtered_mean(vec)
	            ave_slit[izone] = zz[0]
	            ave_slit_sig[izone] = zz[1] > 0.1   ; to trap the central zero dispersion zone
	        endif else begin
	           ave_slit[izone] = 0.0
	           ave_slit_sig[izone] = 1.0
	        endelse
	    endif else begin
	    	ave_slit[izone] = 0.0
	    	ave_slit_sig[izone] = 1.0
	    endelse
    endfor
    ;
    ;   and the mean and dispersion of the deviations
    ;
    ind_ave_slit_sig = WHERE(ave_slit_sig[*] ne 0.0,s_count)
    if(s_count gt 3) then begin
	    zz = median_filtered_mean(ave_slit_sig)
	    mean_slit_sig = zz[0]
	endif else begin
		mean_slit_sig = 1.0
	endelse
    ;
    ;   find the window
    ;
    slit_window_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_Slit_shape_plot_box")
    WIDGET_CONTROL, slit_window_id, GET_VALUE=window_no
    ;
    device, decomposed=0
    loadct,0, /SILENT
    tvlct, [0, 255, 0, 0, 255], [0, 0, 255, 0, 255], [0, 0, 0, 255, 255], 10
    ;
    wset, window_no
    ;
    ;   set plot width as smallest of: (1) fit ranges; smallest of ((2) average ranges; (3) 5 X ave sigma of fits)
    ;
    ; (1) fit ranges
    ;
    xmax1 = max(c_model)
    xmin1 = min(c_model)
    dx1 = xmax1 - xmin1

    ;
    ; (2) range of average
    ;
    xmax2 = max(ave_slit)+2.0*mean_slit_sig
    xmin2 = min(ave_slit)-2.0*mean_slit_sig
    dx2 = xmax2 - xmin2
    ;
    ; (3) 4 X ave sigma
    ;
    ;dx3 = 10.0 * mean_slit_sig  ; don't let plot get wider than +/- 10 ave sigma
    ;xmin3 = -0.5* dx3
    ;xmax3 = 0.5* dx3
    ;
    ;   test
    ;
    dx = min([dx1, dx2])
    ;
    if(dx eq dx1) then begin
        xmin = xmin1
        xmax = xmax1
    endif
    if(dx eq dx2) then begin
        xmin = xmin2
        xmax = xmax2
    endif
    ;
    ;if(dx eq dx3) then begin
    ;    xmin = xmin3
    ;    xmax = xmax3
    ;endif
    ;
    xmax = xmax + 0.1*dx
    xmin = xmin - 0.1*dx
    dx = xmax - xmin
    ;
    ymin = row_value[0]
    ymax = row_value[nzones-1]
    dy = ymax-ymin
    ymin = ymin-0.05*dy
    ymax = ymax+0.05*dy
    dy = ymax-ymin
    ;
    ; if xy small, reset x tick label format
    ;
    if(dx le 4.0) then begin
        xtf = '(f4.1)'
    endif else begin
        xtf = '(f4.0)'
    endelse
    ;
    plot, ave_slit, row_value, /NODATA, xrange=[xmin, xmax], yrange=[ymin, ymax],$
           xstyle=1, ystyle=1, xtitle="Col Offset", ytitle="Row No.", $
           title="Slit Shapes at Line Centers", xtickformat=xtf
    for iline = 0, nlines-1 do begin
        if(line_good_flag[iline] eq 1) then begin
	        x = REFORM(c_model[*,iline])
	        oplot, x, row_value
	    endif
    endfor
    ;
    ;   and the average
    ;
    oplot, ave_slit, row_value, color=12, thick=2
    ;
    ;     and errors
    ;
    xlow = ave_slit - ave_slit_sig
    xhigh = ave_slit + ave_slit_sig
    for izone = 0, nzones-1 do begin
        yy = [row_value[izone], row_value[izone]]
        oplot, [xlow[izone], xhigh[izone]], yy, color=12
    endfor
    ;
    ;   identify any line with residuals averaging more than 3 sigma out
    ;
    model_residuals = c_model * 0.0
    for izone = 0, nzones-1 do begin
        for iline = 0, nlines-1 do begin
           if(line_good_flag[iline] eq 1) then begin
           	model_residuals[izone, iline] = ABS((c_model[izone, iline] - ave_slit[izone]) / ave_slit_sig[izone])
           endif else model_residuals[izone, iline] = deleted_data
        endfor
    endfor
    ;
    slit_qual = fltarr(nlines)
    ;
    for iline = 0, nlines-1 do begin
    	if(line_good_flag[iline] eq 1) then begin
	        vec = REFORM(model_residuals[*,iline])
	        ;print, "for iline = ",iline, " vec = ",vec
	        zz = median_filtered_mean(vec)
	        ;print,"for iline = ",iline," zz = ",zz
	        slit_qual[iline] = zz[0]
	    endif else begin
	    	slit_qual[iline] = -99.0
	    endelse
    endfor
    ;
    ;   go back and color red the bad slits
    ;
    ave_deviation_limit = 2.0
    ;
    for iline = 0, nlines-1 do begin
    	if(line_good_flag[iline] eq 1) then begin
	        if(slit_qual[iline] gt ave_deviation_limit) then begin
	            WCT_print_text, Event, $
	                 STRING(format='("Bad slit model for line = ",i4," wave = ",F10.6," ave deviation = ",F10.2)',$
	                 iline+1, wave_line[iline],slit_qual[iline])
	           xx = REFORM(c_model[*,iline])
	           oplot, xx, row_value, color=11
	           ;
	           line_good_flag[iline] = 0
	           ;
	        endif else begin
	            ;print,"Good slit model for line ",iline," wave = ",wave_line[iline]," mean deviation = ",slit_qual[iline]
	        endelse
	    endif
    endfor
    ;
    ;--------------------------------- determine 2D dispersion equation -------------
    ;
    ;   1. find dispersion constants as function of row zone
    ;   2. fit those constants for row zone dependence
    ;
    a_0 = dblarr(nzones)
    a_1 = dblarr(nzones)
    a_2 = dblarr(nzones)
    ;
    sa_0 = dblarr(nzones)
    sa_1 = dblarr(nzones)
    sa_2 = dblarr(nzones)
    ;
    c_0 = REFORM(row_fit[0,*])
    c_1 = REFORM(row_fit[1,*])
    c_2 = REFORM(row_fit[2,*])
    ;
    for izone = 0, nzones-1 do begin
        ;
        ;  from the good fits, build the vectors of wavelength
        ;
        ind = WHERE(slit_qual le ave_deviation_limit and line_good_flag eq 1, count)
        if(count gt 0) then begin
           y = wave_line[ind]
           x = c_0[ind] + c_1[ind]*row_value[izone] + c_2[ind]*row_value[izone]*row_value[izone]
           ;
;           if(WCT_struct.instrument eq 0 and WCT_struct.slit eq 4) then begin
;           		a = LINFIT(x,y,SIGMA=s_a,/DOUBLE)
;           		a = [a, 0.0]
;           		s_a = [s_a, 0.0]
;           endif else begin
                a = POLY_FIT(x,y,2, SIGMA=s_a, /DOUBLE, STATUS=xstatus)
;           endelse
           ;
           a_0[izone] = a[0]
           a_1[izone] = a[1]
           a_2[izone] = a[2]
           ;
           sa_0[izone] = s_a[0]
           sa_1[izone] = s_a[1]
           sa_2[izone] = s_a[2]
           ;
        endif else begin
           a_0[izone] = 0.
           a_1[izone] = 0.
           a_2[izone] = 0.
           ;
           sa_0[izone] = 1.0e6
           sa_1[izone] = 1.0e6
           sa_2[izone] = 1.0e6
        endelse
        ;
;        print, "for zone ",izone," for mean row number = ",row_value[izone]
;        for ic = 1,count-1 do begin
;        	print, "Wavelength = ",y[ic]," column number = ",x[ic]
;        endfor
;        print, "  dispersion fit is: wave = ",a[0],' + ',a[1],'*col + ',a[2],'*col*col'
        ;
    endfor
    ;
    ;   now invert against row as the independent variable
    ;
    if(WCT_struct.instrument eq 0 and WCT_struct.slit eq 4) then begin
    	b = [LINFIT(row_value, a_0, MEASURE_ERRORS=sa_0, /DOUBLE),0.0]
		c = [LINFIT(row_value, a_1, MEASURE_ERRORS=sa_1, /DOUBLE),0.0]
		d = [LINFIT(row_value, a_2, MEASURE_ERRORS=sa_2, /DOUBLE),0.0]
	endif else begin
	    b = POLY_FIT(row_value, a_0, 2, MEASURE_ERRORS=sa_0, /DOUBLE)
	    c = POLY_FIT(row_value, a_1, 2, MEASURE_ERRORS=sa_1, /DOUBLE)
	    d = POLY_FIT(row_value, a_2, 2, MEASURE_ERRORS=sa_2, /DOUBLE)
	endelse
    ;
    if(idata gt 1) then begin
        ;
        ;   plot models
        ;
        model_0 = b[0] + b[1]*row_value + b[2]*row_value*row_value
        model_1 = c[0] + c[1]*row_value + c[2]*row_value*row_value
        model_2 = d[0] + d[1]*row_value + d[2]*row_value*row_value
        ;
        window, 13
        ymax = max(a_0)
        ymin = min(a_0)
        dy = ymax-ymin
        ymax = ymax+0.1*dy
        ymin = ymin-0.1*dy
        ;
        plot, row_value, a_0, title="offset term", psym=4, yrange=[ymin, ymax], ystyle=1,$
            xtitle="Row Number",ytitle="Offset Term Value"
        yhigh = a_0+sa_0
        ylow = a_0-sa_0
        errplot, row_value, ylow, yhigh, color=11
        oplot, row_value, model_0, color=12
        ;
        window, 14
        ymax = max(a_1)
        ymin = min(a_1)
        dy = ymax-ymin
        ymax = ymax+0.1*dy
        ymin = ymin-0.1*dy
        ;
        plot, row_value, a_1, title="linear term", psym=4, yrange=[ymin, ymax], ystyle=1,$
            xtitle="Row Number",ytitle="Linear Term Value"
        yhigh=a_1 + sa_1
        ylow = a_1 - sa_1
        errplot, row_value, ylow, yhigh, color=11
        oplot, row_value, model_1, color=12
        ;
        window, 15
        ymax = max(a_2)
        ymin = min(a_2)
        dy = ymax-ymin
        ymax = ymax+0.1*dy
        ymin = ymin-0.1*dy
        ;
        plot, row_value, a_2, title="parabolic term", psym=4, yrange=[ymin, ymax], ystyle=1,$
            xtitle="Row Number",ytitle="Parabolic Term Value"
        yhigh = a_2 + sa_2
        ylow = a_2 - sa_2
        errplot, row_value, ylow, yhigh, color=11
        oplot, row_value, model_2, color=12
        ;
        wait, 0.5
        cursor, xx, yy, /wait
        wdelete, 13
        wdelete, 14
        wdelete, 15
    endif
    ;
    ;   and put into correct text windows
    ;
    a0_b0_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A0_B0_text_box")
    a0_b1_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A0_B1_text_box")
    a0_b2_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A0_B2_text_box")
    WIDGET_CONTROL, a0_b0_box_id, SET_VALUE=STRING(format='(F9.7)',b[0])
    WIDGET_CONTROL, a0_b1_box_id, SET_VALUE=STRING(format='(E13.5)',b[1])
    WIDGET_CONTROL, a0_b2_box_id, SET_VALUE=STRING(format='(E13.5)',b[2])
    ;
    a1_b0_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A1_B0_text_box")
    a1_b1_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A1_B1_text_box")
    a1_b2_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A1_B2_text_box")
    WIDGET_CONTROL, a1_b0_box_id, SET_VALUE=STRING(format='(E13.5)',c[0])
    WIDGET_CONTROL, a1_b1_box_id, SET_VALUE=STRING(format='(E13.5)',c[1])
    WIDGET_CONTROL, a1_b2_box_id, SET_VALUE=STRING(format='(E13.5)',c[2])
    ;
    a2_b0_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A2_B0_text_box")
    a2_b1_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A2_B1_text_box")
    a2_b2_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A2_B2_text_box")
    WIDGET_CONTROL, a2_b0_box_id, SET_VALUE=STRING(format='(E13.5)',d[0])
    WIDGET_CONTROL, a2_b1_box_id, SET_VALUE=STRING(format='(E13.5)',d[1])
    WIDGET_CONTROL, a2_b2_box_id, SET_VALUE=STRING(format='(E13.5)',d[2])
    ;
    ;   put into structure
    ;
    WCT_struct.disp_2D[0,0:2] = b[0:2]
    WCT_struct.disp_2D[1,0:2] = c[0:2]
    WCT_struct.disp_2D[2,0:2] = d[0:2]
    ;
    ; put the structure away
    ;
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
end

;--------
pro S5_Export_On, Event
    ;
    ;   save 2D dispersion fit parameters to disk file
    ;
    ;     start by looking to output_directory
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    output_dir = WCT_struct.output_dir
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    if(output_dir ne '') then begin
        file_out = dialog_pickfile(title="Name of Export Calibration File", path=output_dir)
    endif else begin
        file_out = dialog_pickfile(title="Name of Export Calibration File")
    endelse
    ;
    if(file_out ne '') then begin
        openw, lu_out, file_out, /GET_LUN
        ;
        ;  open structure
        ;
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        ;
        ;  write file header
        ;
        printf, lu_out, ';'
        printf, lu_out, ';  Mimir Spectroscopy Package: Wavelength Calibration Tool'
        printf, lu_out, ';   Wavelength Calibration file'
        printf, lu_out, ';'
        printf, lu_out, ';     This file contains the polynomial coefficients that will'
        printf, lu_out, ';     convert image pixel column and row numbers into wavelength'
        printf, lu_out, ';'
        date_time = SYSTIME()
        printf, lu_out, ';     Date/Time written =  ',date_time
        printf, lu_out, ';'
        printf, lu_out, ';     Cal_mode = ',WCT_struct.cal_mode,' where 0 = Argon, 1 = OH sky, 2 = LM sky'
        printf, lu_out, ';     Spec_mode = ',WCT_struct.spec_mode,' where 0=JHK, 1=K, 2=J-3rd, 3=J-4th, 4=LM, 5=HK, 6=K17, 7=IZYJ-1, 8=IZYJ-2'
        printf, lu_out, ';'
        printf, lu_out, ';-------------------------------------------------------------------------------'
        printf, lu_out, ';'
        printf, lu_out, ';  --> caveat emptor! <---  the row and column numbers are relative to whatever'
        printf, lu_out, ';                           trimming the original Argon or OH image had done to them'
        printf, lu_out, ';'
        printf, lu_out, STRING(format='(I4.4,26x,";TRIM_COL")',WCT_struct.trim_col)
        printf, lu_out, STRING(format='(i4.4,26x,";TRIM_ROW")',WCT_struct.trim_row)
        printf, lu_out, STRING(format='(i4.4,26x,";NCOL")',WCT_struct.ncol_result)
        printf, lu_out, STRING(format='(i4.4,26x,";NROW")',WCT_struct.nrow_result)
        printf, lu_out, ';'
        printf, lu_out, ';--------------------------------------------------------------------------------
        printf, lu_out, ';  Wavelength = a_0(row) + a_1(row)*column + a_2(row)*column^2 '
        printf, lu_out, ';'
        printf, lu_out, ';    a_0(row) = b_0 + b_1 * row + b_2 * row^2'
        printf, lu_out, ';    a_1(row) = c_0 + c_1 * row + c_2 * row^2'
        printf, lu_out, ';    a_2(row) = d_0 + d_1 * row + d_2 * row^2'
        printf, lu_out, ';'
        printf, lu_out, ';  so the order of operation is to use the image pixel row number to determine'
        printf, lu_out, ';    the a_0, a_1, and a_2 coefficients from the b, c, d set of coefficients,'
        printf, lu_out, ';    then, use the a_ coefficients with the pixel column number to determine'
        printf, lu_out, ';    the wavelength of the center of the pixel'
        printf, lu_out, ';'
        printf, lu_out, ';  the coefficients are presented below as a row of b, then c, then d coefficients'
        printf, lu_out, ';    with the _0 term first, followed by the _1 and _2 terms, respectively'
        printf, lu_out, ';'
        printf, lu_out, STRING(format='(E20.9,5x,E20.9,5x,E20.9,10x,";B Coeffs")',WCT_struct.disp_2D[0,0],$
                            WCT_struct.disp_2D[0,1],WCT_struct.disp_2D[0,2])
        printf, lu_out, STRING(format='(E20.9,5x,E20.9,5x,E20.9,10x,";C Coeffs")',WCT_struct.disp_2D[1,0],$
                            WCT_struct.disp_2D[1,1],WCT_struct.disp_2D[1,2])
        printf, lu_out, STRING(format='(E20.9,5x,E20.9,5x,E20.9,10x,";D Coeffs")',WCT_struct.disp_2D[2,0],$
                            WCT_struct.disp_2D[2,1],WCT_struct.disp_2D[2,2])
        ;
        ;   close the file
        ;
        FREE_LUN, lu_out
        WCT_print_text, Event, "Calibration File Exported to Disk"
        ;
        WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
        ;
    endif else begin
      WCT_print_text, Event, "Operation Cancel Detected. Continuing."
    endelse
end

;---------
pro S5_Import_On, Event
    ;
    ;  imports a previously written calibration file
    ;
    file_in = dialog_pickfile(title="Calibration File to Import")
    if(file_in ne '') then begin
        openr, lu_in, file_in, /GET_LUN
        ;
        ;  open structure
        ;
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        ;
        ; read the lines, skipping any with a semicolon in the first character
        ;
        file_str = strarr(20)
        istr = 0
        ;
        WHILE ~EOF(lu_in) do begin
            line = ''
            readf, lu_in, line
            ;
            ;  test for semicolon in first column
            ;
            if(~STRCMP(line, ';', 1)) then begin
               file_str[istr] = line
               istr++
            endif
        ENDWHILE
        FREE_LUN, lu_in
        ;
        ;   echo as debug
        ;
        nstr = istr
        file_str = file_str[0:nstr-1]
        ;
        for istr = 0, nstr-1 do begin
            ;print, "input line ",istr," => ",file_str(istr)
            ;;
            ;  look for keywords
            ;
            st = file_str[istr]
            ;
            if(STRPOS(st, "TRIM_COL") ne -1) then reads, st, format='(i4)', trim_col
            if(STRPOS(st, "TRIM_ROW") ne -1) then reads, st, format='(i4)', trim_row
            if(STRPOS(st, "NCOL") ne -1) then reads, st, format='(i4)', ncol
            if(STRPOS(st, "NROW") ne -1) then reads, st, format='(i4)', nrow
            if(STRPOS(st, "B Coeffs") ne -1) then reads, st, format='(3(e20.9,5x))',b_0, b_1, b_2
            if(STRPOS(st, "C Coeffs") ne -1) then reads, st, format='(3(e20.9,5x))',c_0, c_1, c_2
            if(STRPOS(st, "D Coeffs") ne -1) then reads, st, format='(3(e20.9,5x))',d_0, d_1, d_2
            ;
        endfor
       ;
       ;print, "Read in Trim Col, Row = ",trim_col, trim_row
       ;print, "Read in Ncol, Nrow = ",ncol, nrow
       ;
       ; save into WCT structure
       ;
       WCT_struct.trim_col = trim_col
       WCT_struct.trim_row = trim_row
       WCT_struct.ncol_result = ncol
       WCT_struct.nrow_result = nrow
       ;
       WCT_struct.disp_2D[0,0:2] = [b_0, b_1, b_2]
       WCT_struct.disp_2D[1,0:2] = [c_0, c_1, c_2]
       WCT_struct.disp_2D[2,0:2] = [d_0, d_1, d_2]
       ;
       ; now, echo to the display buffer on the step 5 page
       ;
       ;   and put into correct text windows
        ;
        a0_b0_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A0_B0_text_box")
        a0_b1_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A0_B1_text_box")
        a0_b2_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A0_B2_text_box")
        WIDGET_CONTROL, a0_b0_box_id, SET_VALUE=STRING(format='(F9.7)',b_0)
        WIDGET_CONTROL, a0_b1_box_id, SET_VALUE=STRING(format='(E13.5)',b_1)
        WIDGET_CONTROL, a0_b2_box_id, SET_VALUE=STRING(format='(E13.5)',b_2)
        ;
        a1_b0_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A1_B0_text_box")
        a1_b1_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A1_B1_text_box")
        a1_b2_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A1_B2_text_box")
        WIDGET_CONTROL, a1_b0_box_id, SET_VALUE=STRING(format='(E13.5)',c_0)
        WIDGET_CONTROL, a1_b1_box_id, SET_VALUE=STRING(format='(E13.5)',c_1)
        WIDGET_CONTROL, a1_b2_box_id, SET_VALUE=STRING(format='(E13.5)',c_2)
        ;
        a2_b0_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A2_B0_text_box")
        a2_b1_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A2_B1_text_box")
        a2_b2_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S5_A2_B2_text_box")
        WIDGET_CONTROL, a2_b0_box_id, SET_VALUE=STRING(format='(E13.5)',d_0)
        WIDGET_CONTROL, a2_b1_box_id, SET_VALUE=STRING(format='(E13.5)',d_1)
        WIDGET_CONTROL, a2_b2_box_id, SET_VALUE=STRING(format='(E13.5)',d_2)
        ;
        WCT_print_text, Event, "Successfully Imported Calibration File"
        ;
       WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
   endif else begin
        WCT_print_text, Event, "Cancel Operation Detected. Continuing"
   endelse
end


;-----------------------------------------------------------------
pro S5_Done, Event
    WCT_print_text, Event, "Closing Tab 5, Opening Tab 6"
    ;
    tab_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="Step_tab")
    WIDGET_CONTROL, tab_ID, SET_TAB_CURRENT=6
    New_Tab, Event
end

;-----------------------------------------------------------------

pro S6_Select_Images, Event
    ;
    ;  select input images
    ;  based on current input directory, if present
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    input_dir = WCT_struct.input_dir
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    WCT_print_text, Event, "Select Input Images"
    if(input_dir ne '') then begin
        files = dialog_pickfile(TItle="Select Input Images",$
        /MULTIPLE_FILES, path=input_dir)
    endif else begin
        files = dialog_pickfile(TItle="Select Input Images",$
        /MULTIPLE_FILES)
    endelse
    ;
    nfiles = N_ELEMENTS(files)
    ;
    if(nfiles ne 0) then begin
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        WCT_struct.nfiles = nfiles
        WCT_struct.files_multi[0:nfiles-1] = files
        WCT_struct.files_multi[nfiles:*] = ''
        WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
        ;
        WCT_print_text, Event, STRING(format='("Selected ",i4," science images")',nfiles)
    endif else begin
       WCT_print_text, Event, "Cancel Operation Detected. Continuing"
    endelse
end

;-----------------------------------------------------------------
pro S6_Select_Input_Directory, Event
    ;
    ;  based on current input directory, if present
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    input_dir = WCT_struct.input_dir
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    WCT_print_text, Event, "Setting Input Directory"
    if(input_dir ne '') then begin
        dir_in = dialog_pickfile(TItle="Select Input Directory",$
        /directory, path=input_dir)
    endif else begin
        dir_in = dialog_pickfile(TItle="Select Input Directory",$
        /directory)
    endelse
    ;
    if(dir_in ne '') then begin
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        WCT_struct.input_dir = dir_in
        ;print, "Input Directory = ",dir_in
        WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
        ;
        text_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S6_Input_Directory_text_box")
        ;print,"input text widget id = ",text_ID
        WIDGET_CONTROL, text_ID, SET_VALUE=dir_in
    endif else begin
       WCT_print_text, Event, "Cancel Operation Detected. Continuing"
    endelse
end

;-----------------------------------------------------------------
pro S6_Select_Output_On, Event
    ;
    ;  based on current output directory, if present
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    output_dir = WCT_struct.output_dir
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    WCT_print_text, Event, "Setting Output Directory"
    if(output_dir ne '') then begin
        dir_out = dialog_pickfile(TItle="Select Output Directory",$
        /directory, path=output_dir)
    endif else begin
        dir_out = dialog_pickfile(TItle="Select Output Directory",$
        /directory)
    endelse
    ;
    if(dir_out ne '') then begin
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        WCT_struct.output_dir = dir_out
        ;print, "Input Directory = ",dir_in
        WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
        ;
        text_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S6_Output_Directory_text_box")
        ;print,"input text widget id = ",text_ID
        WIDGET_CONTROL, text_ID, SET_VALUE=dir_out
    endif else begin
       WCT_print_text, Event, "Cancel Operation Detected. Continuing"
    endelse
end

;-----------------------------------------------------------------
pro S6_Create_Output_On, Event
    ;
    ;  start with guess of old output directory
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    output_dir = WCT_struct.output_dir
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    ;  create the output directory and set path to it
    ;
    WCT_print_text, Event, "Use file chooser to create new directory"
    ;
    dummy = 'dummy'
    ;
    if(output_dir ne '') then begin
        file_dummy = dialog_pickfile(title="Create New Output Directory", $
           FILE=dummy, PATH=output_dir)
    endif else begin
        file_dummy = dialog_pickfile(title="Create New Output Directory", $
           FILE=dummy)
    endelse
    ;
    if(file_dummy ne '') then begin
        ;
        ;  trim pathname to get just the directory path
        ;
        output_dir = FILE_DIRNAME(file_dummy, /MARK_DIRECTORY)
        ;
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        WCT_struct.output_dir = output_dir
        FILE_MKDIR, output_dir
        WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
        ;
        text_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="S6_Output_Directory_text_box")
        WIDGET_CONTROL, text_ID, SET_VALUE=output_dir
    endif else begin
       WCT_print_text, Event, "Cancel Operation Detected. Continuing"
    endelse
end

;-----------------------------------------------------------------
pro S6_Suffix_text_box, Event
    ;
    ;  text has been entered into the S6 Suffix text box
    ;
    ;  grab it, change the Suffix buffer
    ;
    text_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S6_Suffix_text_box")
    ;
    WIDGET_CONTROL, text_box_id, GET_VALUE=text_string
    ;
    ;print, "S5 Zone size text box string = ",text_string
    ;
    ;  trim and interpret
    ;
    text_string = STRTRIM(text_string,2)
    suffix = ''
    reads, text_string, suffix
    ;
    ;print, "new zone size = ",new_zone_size
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    WCT_struct.suffix = suffix
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    WCT_print_text, Event, STRING(format='("Suffix Changed to ",a)',suffix)
end
;
;-----------------------------------------------------------------
pro S6_Oversample_text_box, Event
    ;
    ;  text has been entered into the S6 Oversample text box
    ;
    ;  grab it, change the Oversample buffer
    ;
    text_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S6_Oversample_text_box")
    ;
    WIDGET_CONTROL, text_box_id, GET_VALUE=text_string
    ;
    ;  trim and interpret
    ;
    text_string = STRTRIM(text_string,2)
    reads, text_string, oversample
    ;
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    if(oversample gt 0 and oversample lt 10) then begin
        WCT_struct.oversample = oversample
    endif
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    WCT_print_text, Event, STRING(format='("Oversample Factor Changed to ",i)',oversample)
end
;--------------------------------
pro S6_Start_Correction_On, Event
    ;
    ;  start the loop of reading science images, trimming, wavelength
    ;  calibrating, and saving (with updated FITS headers)
    ;
    ;  1. from the number of images to process, build the cleared LED stack
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    ;
    ;   debug
    ;
    idata = WCT_struct.debug_flag   ; set to 0 for no plots
    ;
    nfiles = WCT_struct.nfiles
    ;
    oversample = double(WCT_struct.oversample)
    ;
    ;   test for no files to correct
    ;
    if(nfiles le 0) then begin
        WCT_print_text, Event, "No files selected. Please try again..."
        goto, fast_exit
    endif
    ;
    ; clear the LED stack to gray
    ;
    bar_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S6_Progress_bar")
    colors = strarr(nfiles)
    for ifile = 0, nfiles-1 do begin
        colors[ifile] = 'gray'
    endfor
    WCT_LED_stack, bar_id, colors
    ;
    ;   do a bunch of set up
    ;
    output_dir = WCT_struct.output_dir
    suffix = WCT_struct.suffix
    ;
    spec_mode = WCT_struct.spec_mode
    ;
    ch_1 = WCT_struct.wave_limits[spec_mode, 0]
    ch_2 = WCT_struct.wave_limits[spec_mode, 1]
    ;
    case spec_mode of
          -1:   begin
                 WCT_print_text, Event, "ERROR - No Spec Mode Selected. Go to Step 1"
                 goto, fast_exit
             end
           0:    begin           ; JHK-Spec
                 ;
                 dvch_i = 0.00005d0     ; for interpolated spectrum, not final
                 dvch_f = 0.0014d0       ; for final spectrum
             end
           1:    begin            ; K-Spec
                  dvch_i = 0.00005d0
                  dvch_f = 0.0014d0
             end
           2:    begin           ; J-Spec - 3rd order
                  dvch_i = 0.00005d0
                  dvch_f = 0.00125d0
             end
           3: begin
                  dvch_i = 0.00003d0  ; J-Spec - 4th order
                  dvch_f = 0.00085d0
              end
           4: begin
                   dvch_i = 0.0002d0
                   dvch_f = 0.0032d0   ; LM TBD
              end
           5: begin           ; HK-Spec
                 ;
                 dvch_i = 0.00005d0     ; for interpolated spectrum, not final
                 dvch_f = 0.0014d0       ; for final spectrum
              end
           6: begin				; K17 Spec
           		;
           		dvch_i = 0.00002d0
           		dvch_f = 0.00050d0
           	  end
           7: begin				; IZYJ-1st
           		dvch_i = 0.00005d0     ; for interpolated spectrum, not final
                dvch_f = 0.00140d0       ; for final spectrum
              end
           8: begin				; IZYJ-2nd
           		dvch_i = 0.000025d0     ; for interpolated spectrum, not final
                dvch_f = 0.000700d0       ; for final spectrum
              end
           9: begin				; H17 Spec
           		;
           		dvch_i = 0.00002d0
           		dvch_f = 0.0005d0
           	  end
           10: begin				; IZ17 Spec
           		;
           		dvch_i = 0.00001d0
           		dvch_f = 0.00025d0
           	  end
           10: begin
                   dvch_i = 0.00005d0   ; BVR/VPH TBD
                   dvch_f = 0.00100d0
              end
    endcase
    ;
    ;   include oversampling
    ;
    dvch_f = dvch_f / double(oversample)
    ;
    files = WCT_struct.files_multi[0:nfiles-1]
    ;
    ncol_result = WCT_struct.ncol_result
    nrow_result = WCT_struct.nrow_result
    ;
    if(ncol_result eq 0) then ncol_result = WCT_struct.ncol_raw
    if(nrow_result eq 0) then nrow_result = WCT_struct.nrow_raw
    ;
    trim_col = WCT_struct.trim_col
    trim_row = WCT_struct.trim_row
    ;
    ; get the 2D dispersion equation
    ;
    b = WCT_struct.disp_2D[0,*]
    ;print, "b = ",b
    c = WCT_struct.disp_2D[1,*]
    ;print, "c = ",c
    d = WCT_struct.disp_2D[2,*]
    ;print, "d = ",d
    ;
    x_col = dindgen(ncol_result) + 0.5
    ;
    ;print, "x_col = ",x_col
    ;
    ;print, "number of columns, rows = ",ncol_result, nrow_result
    ;
    ;-------- interpolated spectrum --------------
    ;
    nch_i = ROUND((ch_2 - ch_1) / dvch_i) + 1
    x_i_col = dindgen(nch_i) + 0.5d0
    new_wave = ch_1 + (x_i_col - 0.5)*dvch_i           ; wavelength index for interpolated spectrum
    ;
    new_spec = dblarr(nch_i)                         ; interpolated spectrum for one row of data
    ;
    ;-------- and resampled spectral image -----
    ;
    nch_f = ROUND((ch_2 - ch_1) / dvch_f) + 1
    x_f_col = dindgen(nch_f) + 0.5d0
    wave_f = ch_1 + (x_f_col - 0.5d0) * dvch_f          ; final spectrum - referenced to pixel centers (0.5)
    ;
    resampled_spec = dblarr(nch_f,nrow_result)          ; new image to hold resampled image
    ;
    ratio = ROUND(dvch_f/dvch_i)
    ;
    if(nfiles gt 0) then begin
       ;
       ; For each Science Image,
       ;
       for ifile = 0, nfiles-1 do begin
         ;
         ;  hourglass
         ;
         WIDGET_CONTROL, /HOURGLASS
         ;
         colors[ifile] = 'red'
         WCT_LED_stack, bar_id, colors
         ;
         ;    read the image
         ;
         science_image = readfits(files[ifile],header,/SILENT)
         ind_nan = WHERE(~FINITE(science_image), n_nan)
         if(n_nan gt 0) then science_image[ind_nan] = -1.0e6
         WCT_print_text, Event, STRING(format='("Read in science image ",a)',FILE_BASENAME(files[ifile]))
         ;
         colors[ifile] = 'yellow'
         WCT_LED_stack, bar_id, colors
         ;
         ;    apply masking
         ;
         image_size = SIZE(science_image)
         ncol_sci = image_size[1]
         nrow_sci = image_size[2]
         ;
         if(ncol_sci ne ncol_result or nrow_sci ne nrow_result) then begin
          science_image = science_image[trim_col:trim_col+ncol_result-1,trim_row:trim_row+nrow_result-1]
         endif
         ;
         ;  clean
         ;
         WCT_print_text, Event, "Cleaning Bad Pixels"
         science_image = fix_bad_pixels(science_image)
         ;
         ;print, "After cleaning, Image size = ",SIZE(science_image)
         ;
         colors[ifile] = 'blue'
         this_color = 'blue'
         WCT_LED_stack, bar_id, colors
         ;
         ;    calibrate each row
         ;
         di_df = dvch_f / dvch_i		; ratio of channel sizes for final vs interpolated spectra
         ;								; should be ~10
         ;	compute start, stop integration channels for interpolated spectra to be
         ;		averaged onto the resulting spectra
         ;
         int_ch0 = lonarr(nch_f)
         int_ch1 = int_ch0
         int_ch2 = fltarr(nch_f)
         ;
         ich = lindgen(nch_f)
         int_ch0[*] = (ROUND( (FLOAT(ich)-0.5)*di_df) ) > 0L		; start channel
         int_ch1[*] = (int_ch0[*] + CEIL(di_df) -1) < (nch_i - 1)	; end channel
         int_ch2[*] = FLOAT(int_ch1[*] - int_ch0[*]) + 1.0		; number of channels
         ;
         for irow = 0, nrow_result-1 do begin
         	  ;
         	  ;	flash LED light every 50th row
         	  ;
         	  row5 = ROUND(irow / 50.)
         	  if(50*row5 eq irow) then begin
         	  	colors[ifile] = this_color
         	  	if(this_color eq 'orange') then this_color = 'blue' else this_color = 'orange'
         	  	WCT_LED_stack, bar_id, colors
         	  endif
              ;
              spec_line = REFORM(science_image[*,irow])
              ;
              zrow = double(irow)
              ;
              a_0 = b[0] + b[1]*zrow + b[2]*zrow*zrow
              a_1 = c[0] + c[1]*zrow + c[2]*zrow*zrow
              a_2 = d[0] + d[1]*zrow + d[2]*zrow*zrow
              ;
              ;   by calibrating the x (column) axis
              ;
              x_wave = a_0 + a_1 * x_col + a_2 * x_col * x_col
              ;
              ;   Interpolate, correct for changing channel width
              ;
              corr = dvch_f / (a_1 + 2.0d0 * a_2 * x_col)   ; bandwidth correction
              spec_line = spec_line * corr
              ;
              spec_line_f = FLOAT(spec_line)
		  	  ;
		  	  ;	first, compute new indices for new_wave
		  	  ;
		  	  ;new_i = FLOAT((1./(2.0*a_2)) * (-1.0*a_1 + sqrt( (a_1 * a_1) - 4.0 * (a_0 - new_wave) * a_2)))
			  ;
              	;
              	;	the original way...
              	;
              	new_spec = INTERPOL(spec_line, x_wave, new_wave)     ; no flag makes this a linear interpolation
				;
				;	the way GPU demands
				;
;				new_spec = INTERPOLATE(spec_line_f, new_i)
				;
                ;   bin by averaging, onto final spectrum wave limits
                ;
                for ii = 0, nch_f-1 do resampled_spec[ii,irow] = TOTAL(new_spec[int_ch0[ii]:int_ch1[ii]])/int_ch2[ii]
                ;

              ;
              if(idata eq 0) then begin
                ;
                if(irow eq 200) then begin
                    device, decomposed=0
                    loadct,0, /SILENT
                    tvlct, [0, 255, 0, 0, 255], [0, 0, 255, 0, 255], [0, 0, 0, 255, 255], 10
                    ;
                    base_window_plot_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="base_plot")
                    WIDGET_CONTROL, base_window_plot_id, GET_VALUE=win_no
                    ;
                    wset, win_no
                    ;window, 12, xsize=1400, ysize=900
                    ind = WHERE(x_wave ge ch_1 and x_wave le ch_2, count)
                    if(count gt 0) then begin
                        zz = median_filtered_mean(spec_line[ind])
                        ;
                        ymin = zz[0]-2.0*zz[1]
                        ;ymax = zz[0]+20.0*zz[1]
                        ;
                        ymax = MAX(spec_line[ind])
                        dy = ymax - ymin
                        ymax = ymax + 0.05 * dy
                        ;
                        plot, x_wave, spec_line, xrange=[ch_1, ch_2], yrange=[ymin, ymax], $
                           xstyle=1, ystyle=1, /NODATA, xtitle="Wavelength [microns]",$
                           ytitle="Counts", title="Row 200 Spectrum: Red=original, Blue=interpolated"
                        oplot, x_wave, spec_line, color=11
                        oplot, wave_f, resampled_spec[*,irow], color=13
                    endif
                endif
              endif

        endfor
        ;
;        end_time = SYSTIME(1)
;        dif_time = end_time - start_time
;        WCT_print_text, Event, "Total Conversion time = "+STRING(dif_time)+" seconds"
        ;
        ;
        ;   update header parameters
        ;
        case WCT_struct.instrument of
        	0:	comment = 'Mimir Instrument'
        	1: 	comment = 'PRISM Instrument'
        endcase
        FXADDPAR, header, 'WCT_INST', WCT_struct.instrument, comment
        ;
        case WCT_struct.camera of
        	0: comment = 'Mimir F/5 Camera'
        	1: comment	= 'Mmir F/17 Camera'
        endcase
        FXADDPAR, header, 'WCT_CAM', WCT_struct.camera, comment
        ;
        case WCT_struct.slit of
        	1: comment = 'S1 - Tall, Narrow Slit'
        	4: comment = 'S4 - Fat-eared, Slit'
        	11: comment = 'S11 - Short, Narrow Slit'
        	else: comment = 'Unknown Slit'
        endcase
    	;
        FXADDPAR, header, "WCT_SLIT", WCT_struct.slit,comment
        ;
        case WCT_struct.object of
        	0: comment = 'Single Object'
        	else: comment = 'Multiple Object'
        endcase
        FXADDPAR, header, 'WCT_OBJ', WCT_struct.object, comment
        ;
        case WCT_struct.spec_mode of
            0: comment = "JHK-Spec Mode"
            1: comment = "K-Spec Mode"
            2: comment = "J-Spec 3rd Order Mode"
            3: comment = "J-Spec 4th Order Mode"
            4: comment = "LM-Spec Mode"
            5: comment = "HK-Spec Mode"
            6: comment = "K17-Spec Mode"
            7: comment = "IZYJ-Spec 1st"
            8: comment = "IZYJ-Spec 2nd"
            9: comment = "H17-Spec Mode"
            10: comment = "IZ17-Spec Mode"
            11: comment = "BVR/VPH-Spec Mode"
        endcase
        FXADDPAR, header, "SPECMODE", WCT_struct.spec_mode,comment
        FXADDPAR, header, 'WCT_SPEC', WCT_struct.spec_mode, comment
        ;
        case WCT_struct.cal_mode of
            0: comment = 'Calibrated using Argon Lamp Lines'
            1: comment = 'Calibrated using OH night sky Lines'
            2: comment = 'Calibrated using LM night sky features'
        endcase
        FXADDPAR, header, "CAL-MODE", WCT_struct.cal_mode, comment
        FXADDPAR, header, 'WCT_CAL', WCT_struct.cal_mode, comment
        ;
        if(WCT_struct.oversample ne 1) then begin
            FXADDPAR, header, "OVERSAMP", WCT_struct.oversample, "Spectral Oversample Factor"
        endif
        FXADDPAR, header, "TRIM-ROW", WCT_struct.trim_row, "Mimir image trim - rows skipped from raw image edge"
        FXADDPAR, header, "TRIM-COL", WCT_struct.trim_col, "Mimir image trim - cols skipped from raw image edge"
        ;
        FXADDPAR, header, "CTYPE1", "WAVE", " Wavelength"
        ;
        FXADDPAR, header, "CRVAL1", double(ch_1),   " Channel 1 Wavelength in microns", FORMAT='(e13.6)'
        ;FXADDPAR, header, "CRDELT1", rdvch, " Wavelength channel width and step"
        FXADDPAR, header, "CUNIT1", "um",   " Microns Wavelength"
        FXADDPAR, header, "CTYPE2", "LINEAR"
        FXADDPAR, header, "CUNIT2", "pixels"
        FXADDPAR, header, "CD1_1", double(dvch_f), FORMAT='(e13.6)'
        FXADDPAR, header, "CD2_2", 1.00d0, FORMAT='(e13.6)'
        FXADDPAR, header, "CD1_2", 0.0d0, FORMAT='(e13.6)'
        FXADDPAR, header, "CD2_1", 0.0d0, FORMAT='(e13.6)'
        SXDELPAR, header, ["CRDELT1 ", "CRDELT2 "]
        ;
        ;  add info about calibration (TBD)
        ;
        FXADDPAR, header, "COMMENT", STRING(format='("Wavelength Calibrated by Mimir SP-WCT on ",a)',SYSTIME())
        FXADDPAR, header, "COMMENT", "Converted to linear dispersion and linear slit using 2D dispersion equation:"
        FXADDPAR, header, "COMMENT", "   lambda = a_0 + a_1 * col + a_2 * col * col"
        FXADDPAR, header, "COMMENT", "      where a_0 = b_0 + b_1 * row + b_2 * row * row"
        FXADDPAR, header, "COMMENT", "            a_1 = c_0 + c_1 * row + c_2 * row * row"
        FXADDPAR, header, "COMMENT", "            a_2 = d_0 + d_1 * row + d_2 * row * row"
        FXADDPAR, header, "MWCTA0B0", b[0]," 2D dispersion b_0 constant"
        FXADDPAR, header, "MWCTA0B1", b[1]," 2D dispersion b_1 constant"
        FXADDPAR, header, "MWCTA0B2", b[2]," 2D dispersion b_2 constant"
        FXADDPAR, header, "MWCTA1C0", c[0]," 2D dispersion c_0 constant"
       FXADDPAR, header, "MWCTA1C1", c[1]," 2D dispersion c_1 constant"
       FXADDPAR, header, "MWCTA1C2", c[2]," 2D dispersion c_2 constant"
       FXADDPAR, header, "MWCTA2D0", d[0]," 2D dispersion d_0 constant"
       FXADDPAR, header, "MWCTA2D1", d[1]," 2D dispersion d_1 constant"
       FXADDPAR, header, "MWCTA2D2", d[2]," 2D dispersion d_2 constant"
        ;---------------
        ;
        ;  create output file name
        ;
        out_name = output_dir + FILE_BASENAME(files[ifile],".fits",/FOLD_CASE) + suffix + ".fits"
        FXADDPAR, header, "FILENAME", FILE_BASENAME(out_name)
        ;
        writefits, out_name, resampled_spec, header
       ;
       WCT_print_text, Event, STRING(format='("Wrote Output Image = ",a)',out_name)
       colors[ifile] = 'green'
       WCT_LED_stack, bar_id, colors
       ;
    endfor
  endif else begin
       WCT_print_text, Event, "No Files Selected for Calibration. Ending"
  endelse
;
fast_exit:
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
end

;-----------------------------------------------------------------
pro S6_Done, Event
    WCT_print_text, Event, "Closing Tab 6, Opening Tab 0"
    ;
    tab_ID = WIDGET_INFO( Event.top, FIND_BY_UNAME="Step_tab")
    WIDGET_CONTROL, tab_ID, SET_TAB_CURRENT=0
    New_Tab, Event
end
;-------------------------------------------------------------
pro Status_Debug_Button_On, Event
    ;
    ;  toggle debug flag and reload status line
    ;
    WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
    ;
    if(WCT_struct.debug_flag eq 0) then WCT_struct.debug_flag = 1 else WCT_struct.debug_flag = 0
    ;
    WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    ;
    WCT_show_status, Event
end
;-----------------------------------------------------------------
; Activate Button Callback Procedure.
; Argument:
;   Event structure:
;
;   {WIDGET_BUTTON, ID:0L, TOP:0L, HANDLER:0L, SELECT:0}
;
;   ID is the widget ID of the component generating the event. TOP is
;       the widget ID of the top level widget containing ID. HANDLER
;       contains the widget ID of the widget associated with the
;       handler routine.

;   SELECT is set to 1 if the button was set, and 0 if released.
;       Normal buttons do not generate events when released, so
;       SELECT will always be 1. However, toggle buttons (created by
;       parenting a button to an exclusive or non-exclusive base)
;       return separate events for the set and release actions.

;   Retrieve the IDs of other widgets in the widget hierarchy using
;       id=widget_info(Event.top, FIND_BY_UNAME=name)

;-----------------------------------------------------------------

