pro WCT_find_spectrum, image, spectra_rows, sky_rows, spec_mode, IDATA=idata
;
;   discovers which rows contain spectra and which
;   can be used for sky line determination
;
;   returns flats for each row for spectra and sky - flag = 0 is false
;
;   DPC     20060218    original code
;           20060304    trimmed regions to proper number after discovery
;           20060410    added code to deal with non-discovered spectrum
;           20060807    added iplot keyword
;   MP      20070116    added two inputs to restrict spectrum center to search region
;   DPC     20070824    moved into WCT package - does image search only over columns
;                       indicated by spec_mode
;
;   clean, then sum along rows
;
if(N_ELEMENTS(IDATA) eq 0) then idebug = 0 else idebug = idata
;
threshold = 4.0        ; 4.0 sigma threshold
;
deleted_data = -1.0e6
;
;   which columns to search?
;
case spec_mode of
    0: col_test = [425, 525]		; JHK
    1: col_test = [750, 850]		; K-Spec
    2: col_test = [365, 465]		; J-Spec 3rd
    3: col_test = [700, 800]		; J-Spec 4th
    4: col_test = [358, 557]        ; LM-TBD
    5: col_test = [550, 750]		; HK-Spec
    6: col_test = [100, 700]		; K17
    7: col_test = [45, 154]		    ; IZYJ-1st
    8: col_test = [569,721]		    ; IZYJ-2nd
    9: col_test = [0, 100]			; PRISM-VPH
endcase
;
;clean_image = fix_bad_pixels(image)
clean_image = image
;
npars = N_PARAMS()
;
image_size = SIZE(clean_image)
ncol = image_size[1]
nrow = image_size[2]
;
;
;if(npars lt 7) then begin
    col0 = col_test[0]
    col1 = col_test[1]
    row0 = 0
    row1 = nrow-1
;endif
;
;   compress image
;
clean_image = clean_image[col0:col1,*]
;
row_sum = dblarr(nrow)
;
for ir = 0, nrow-1 do begin
    row_sum[ir] = 0.0d0
    vec = REFORM(clean_image[*,ir])
    ;
    ind = WHERE(vec gt -250. and vec lt 300000.,icount)
    if(icount gt 5) then begin
    	r_vec = vec[ind]
    	zz = median_filtered_mean(r_vec)
    endif else zz = [0.,0.]
    if(icount gt 5) then row_sum[ir] = zz[0] else row_sum[ir] = deleted_data
endfor
;
xr = indgen(nrow)
;
;   median filtered mean
;
zz = median_filtered_mean(row_sum)
;
;   plot, if debug set
;
if(idebug eq 1) then begin
    device, decomposed=0
    black = 14
    loadct,0
    tvlct, [0, 255, 0, 0, 255], [0, 0, 255, 0, 255], [0, 0, 0, 255, 255], 10
    window, 10, title="Row Averaged Intensity"
    plot, xr, row_sum, xtitle="Row [pixels]", ytitle="Summed Counts"
    ;
    plots, [xr[0],xr[nrow-1]],[zz[0],zz[0]],color=12
endif
;
spectra_rows = intarr(nrow)     ; will hold list of rows with emission
sky_rows = intarr(nrow)          ; will hold list of rows without emission
;
ind = WHERE((row_sum[*] gt threshold*zz[1]+zz[0]) and row_sum[*] ne deleted_data,nobj)
;
if(idebug eq 1) then print, "Found ",nobj," rows with significant object emission"
if(nobj eq 0) then begin
    if(idebug eq 1) then print, " - No spectrum found"
    ;
    spectra_rows[*] = 0
    sky_rows[*] = 1
endif else begin
    spectra_rows[*] = 0
    sky_rows[*] = 1
    spectra_rows[ind] = 1
    sky_rows[ind] = 0
endelse
;
if(idebug eq 1 and nobj ne 0) then oplot, xr[ind], row_sum[ind],color=13,psym=4
;
if(idebug eq 1) then begin
    wait, 0.5
    cursor, xx, yy, /wait
    wait, 0.5
    wdelete, 10
endif
;
end
