pro WCT_imcombine_no_bad_pix, Event, spec_mode, input_list, output_average_image, output_FITS_header, $
			WCT_struct
;
;   image combiner, ala IRAF imcombine -- this version ignores badpixel masks
;
;   accepts list of images and list of offsets and performs median-filtered-mean
;   averaging, producing an averaged image, a coverage image, and a sigma image
;
;   at least one of the images should have zero offsets
;
;   if SKY switch on, sky values removed before combining, saved to header
;   if LONG_OUTPUT switch on, output standard deviation and coverage images will be stored
;
;   DPC 20041014
;       20041031 - updated for Oct '04 image sizes
;     20051102 modified to allow command line arguments and to change to single file/offsets input file
;           also, if sky removal selected, sky value removed is stored as comment in header
;     20051104 - modified from imcombine
;     20070626 - added query about offsets
;     20070803 - added no sky removal option
;     20070808 - copied from imcombine_no_bad_pix in main Mimir S/W directory, tweaked for
;          use in WCT
;     20070834 - added find spectrum call to locate, remove significant continuum emission
;	  20090318 - modified for fat-eared slit, removed search for continuum for argon
;
;------------------------------------------------------------------------------
SKY=0
MASK=0
;
xraw = 1024
yraw = 1026
;
XL = WCT_struct.trim_col
XR = WCT_struct.ncol_result - 1 + XL
YL = WCT_struct.trim_row
YR = WCT_struct.nrow_result - 1 + YL
;
;case spec_mode of
;    4: begin          ; LM TBD
;         XL = 0
;         XR = 1023
;         YL = 260
;         YR = 760
;       end
;    5: begin          ; PRISM TBD
;         XL = 0
;         XR = 1023
;         YL = 260
;         YR = 760
;       end
;    else:  begin
;         end
;endcase
;
deleted_data = -1.0d6
;
image_list = input_list
;
;   read input list of files, offsets
;
nimages = N_ELEMENTS(image_list)
images = image_list
;
; determine size of output image
;
test = readfits(images[0],header,/SILENT)
ind_nan = WHERE(~FINITE(test),n_nan)
if(n_nan gt 0) then test[ind_nan] = -1.0e6
image_size = SIZE(test)
ncol = image_size[1]
nrow = image_size[2]
;
;   if this was a raw image, turn on masking, else
;   leave masking off
;
if(ncol eq xraw and nrow eq yraw) then begin
    MASK = 1
    MR513 = 1
    xsize = XR - XL + 1
    ysize = YR - YL + 1
endif else begin
    MASK = 0
    MR513 = 0
    xsize = ncol
    ysize = nrow
endelse
;
;   reserve this much space for the output images
;
image_mean = dblarr(xsize, ysize)
;
;   and the monster array to hold all the images
;
big_image = fltarr(nimages, xsize, ysize)
;
;   clear it to all deleted pixels
;
big_image[*,*,*] = deleted_data
;
;   read in all the images into the big_image array
;
if(SKY eq 1) then save_sky = fltarr(nimages)
;
for ifile = 0, nimages-1 do begin
    WCT_print_text, Event, STRING(format='("Reading Image ",a)',images[ifile])
    ;
    raw_image = readfits(images[ifile], header,/SILENT)
    ind_nan = WHERE(~FINITE(raw_image),n_nan)
    if(n_nan gt 0) then raw_image[ind_nan] = -1.0e6
    ;
    ;   if masking of row 513 turned on, set it to all deleted data
    ;
    if(MR513 eq 1) then begin
        raw_image[*,513] = deleted_data
    endif
    ;
    ;   if image masking selected, mask image and update sizes
    ;
    if(MASK eq 1) then begin
        raw_image = raw_image[XL:XR,YL:YR]
    endif
    ;
    this_image_size = SIZE(raw_image)
    this_ncol = this_image_size[1]
    this_nrow = this_image_size[2]
    ;
    ;   save a copy of the first image header as template for output headerr
    ;
    if(ifile eq 0) then begin
        out_header = header
    endif
    ;
    ;     find any spectra and punch them out in each image, if mode not argon
    ;
    if(WCT_struct.cal_mode ne 0) then begin
	    ;
	    ;
	    WCT_print_text, Event, STRING(format='("Testing for Continuum Spectrum in Image ",a)',$
	              FILE_BASENAME(images[ifile]))
	    ;
	    WCT_find_spectrum, raw_image, spec_zone, sky_zone, spec_mode, IDATA=idebug
	    ;
	    spec_count = TOTAL(spec_zone)
	    if(spec_count ne 0) then begin
	        WCT_print_text, Event,"  Spectrum was found -- deleting rows with significant emission"
	        ind = WHERE(spec_zone eq 1)
	        for ispec = 0, spec_count-1 do begin
	           ;
	           ; and set those image rows to deleted data
	           ;
	           raw_image[*,ind[ispec]] = deleted_data
	        endfor
	    endif
	endif
    ;
    ;   pack constituent image into big_image
    ;
    big_image[ifile,*,*] = raw_image
endfor
;
;   for every pixel in the final resulting images...
;
WCT_print_text, Event, "Starting Coaddition Process"
;
p10 = indgen(10)*(CEIL(xsize/10))
v10 = indgen(10)*10
;
for icol = 0, xsize-1 do begin
;
;   print percentage complete to widget
;
    ind_p = WHERE(icol eq p10[*], count)
    if(count ne 0) then begin
       WCT_print_text, Event, STRING(format='("Coaddition is ",i3," percent done")',v10[ind_p])
    endif
;
    for irow = 0, ysize-1 do begin
;
       ;ind = where(big_image[*,icol,irow] ne deleted_data, ngood)
       pix_total = TOTAL(big_image[*,icol,irow])
       pix_num = N_ELEMENTS(big_image[*,icol,irow])
       ;
       if(pix_total lt (-0.75e6)*pix_num) then begin
       		image_mean[icol,irow] = deleted_data ; more than 75% deleted
       endif else begin
       		if(pix_total lt (-0.5e6) * pix_num) then begin
       			ind = WHERE(big_image[*,icol,irow] ne deleted_data, ngood)
       			if(ngood gt 0) then begin
       				 mx = median_filtered_mean(big_image[ind,icol,irow])     ; [0] is mean; [1] is sigma
                     ;
                     ;   if large negative, reset
                     ;
                     if(mx[0] lt -250.0) then begin
                         image_mean[icol,irow] = deleted_data
                     endif else begin
                         image_mean[icol,irow] = mx[0]
                     endelse
                 endif else begin
                 	image_mean[icol,irow] = deleted_data
                 endelse
             endif else begin
             	;
             	;	most of the time come here to do mfm clean
             	;
             	mx = median_filtered_mean(big_image[*,icol,irow])     ; [0] is mean; [1] is sigma
                ;
                ;   if large negative, reset
                ;
                if(mx[0] lt -250.0) then begin
                    image_mean[icol,irow] = deleted_data
                endif else begin
                    image_mean[icol,irow] = mx[0]
                endelse
       		 endelse
       	endelse
    endfor
endfor
;
;   done computing image, update header before writing
;
;   date/time
;   filename
;   method/software used
;   input images, with sky (if SKY=1)
;
stime = SYSTIME()
FXADDPAR, out_header, 'HISTORY',STRING(format='("Produced by WCT_IMCOMBINE on ",a)',stime)
;
scomment = "; No sky removal before averaging"
for ifile = 0, nimages-1 do begin
    FXADDPAR, out_header, "COMMENT",STRING(format='("Included file ",a)',FILE_BASENAME(images[ifile]))+scomment
endfor
;
;   add trim keywords
;
FXADDPAR, out_header, "TRIM-COL", XL, "Cols skipped from raw image edge"
FXADDPAR, out_header, "TRIM-ROW", YL, "Rows skipped from raw image edge"
;
;print, "in subroutine, trim_col, trim_row = ",XL, YL
;
;writefits, out_mean, image_mean, out_header
;
output_average_image = image_mean
output_FITS_header = out_header
;
WCT_print_text, Event, "Image combining is done"
end

