pro WCT_refine_dispersion, Event, input_disp_in, disp_out, wave_fit, x_spec, spec, cal_mode, $
                           lines, line_qual, wave_limits, SAVE_ARRAY=save_array, $
                           IDATA=idata, FIT_WIDTH=fit_width, rms, NOPLOT=noplot, good_fit
;
;   takes an input (channel) spectrum and a nominal
;   dispersion equation and refines the dispersion equation
;   by detailed fitting against nominal line list.
;
;   optional return save array is (3,nlines):
;     0 element is line center wavelength
;     1 element is column fit of center
;     2 element is uncertainty of fit [in columns]
;
;   DPC     20060306    original code
;           2006 Fall   lots of Gully mods
;           20070412    tweaks
;           20070801    modified to work for Ar or OH
;
;==============================================================
;
if(N_ELEMENTS(idata) eq 0) then idata = 0        ; set = 1 for more debug info
if(N_ELEMENTS(NOPLOT) eq 0) then noplot = 0 else noplot = 1
if(N_ELEMENTS(FIT_WIDTH) eq 0) then fit_width=5
;
device, decomposed=0
;
;=============================================================
;
;
if(N_ELEMENTS(save_array) eq 0) then isave = 0 else isave = 1
;
nspec = N_ELEMENTS(spec)
;
;   form list of good lines for fitting
;
;   1. cull bad lines
;
ind = WHERE(line_qual eq 1, count)
if(count gt 0) then begin
    fit_lines = lines[ind]
    ;
    ;  2. check wavelength limits
    ;
    ind2 = WHERE(abs(fit_lines) ge min(wave_limits) and abs(fit_lines) le max(wave_limits),nlines)
    if(nlines gt 0) then begin
       ;
       line_list = fit_lines[ind2]
    endif else begin
       nlines = 0
    endelse
endif else begin
    nlines = 0
endelse
;
;	check for absorption line fits
;
line_abs = intarr(nlines)
ind_abs = WHERE(fit_lines lt 0.0, count)
if(count gt 0) then begin
	line_abs[ind_abs] = 1
	line_list = abs(fit_lines)
endif
;
disp_in = input_disp_in     ; so we don't write over old dispersion equation
;
if(nlines gt 0) then begin
    WCT_print_text, Event, STRING(format='("Fitting to ",i," Calibration Lines")',nlines)
    ;
    ;   plot the spectrum from the start, stop limits and the locations of all the
    ;   cal lines into the base plot region
    ;
    trial_wave = disp_in[0] + disp_in[1]*x_spec + disp_in[2]*x_spec*x_spec
    ind_trial = WHERE(trial_wave ge min(wave_limits) and trial_wave le max(wave_limits),ntrial)
    if(ntrial gt 0) then begin
        trial_wave = trial_wave[ind_trial]
        trial_spec = spec[ind_trial]
        ;
        if(idata gt 0 and noplot eq 0) then begin
            ;
            WCT_plot_spectrum_and_cal_lines, Event, trial_wave, trial_spec, lines, line_qual, $
               cal_mode, TEXT="Before Fitting"
               ;
               print, "Click cursor to continue"
               wait, 0.5
               cursor, xx, yy, /wait
        endif
        ;
    endif
    ;
    ;   check for an open S4 pop up window. If one found, then just update the window numbers, don't kill it
    ;
    if(S4_PU_Exists(Event)) then begin
        WIDGET_CONTROL, Event.top, GET_UVALUE=WCT_struct, /NO_COPY
        top_id = WCT_struct.S4_top_win_no
        mid_id = WCT_struct.S4_middle_win_no
        bot_id = WCT_struct.S4_bottom_win_no
        window_numbers = [top_id, mid_id, bot_id]
        WIDGET_CONTROL, Event.top, SET_UVALUE=WCT_struct, /NO_COPY
    endif else begin
        ;
        ;           if none found, launch the pop up window and return the window numbers
        ;
        S4_Launch_PU, Event, window_numbers, w_ID     ; fire up the pop up window - window numbers are [top, middle, bottom]
        ;
    endelse
    ;
    device, decomposed=0
    loadct,0, /SILENT
    tvlct, [0, 255, 0, 0, 255], [0, 0, 255, 0, 255], [0, 0, 0, 255, 255], 10
    ;
    ncol = N_ELEMENTS(y_spec)
    ;
    if(isave eq 1) then begin
       save_array = dblarr(3,nlines)
       save_array[0,*] = line_list[*]
    endif
    ;
    ;   for each line in the list,
    ;
    fit_col = dblarr(nlines)
    fit_unc = dblarr(nlines)
    ;
    line_col = dblarr(nlines)
    ;
    ;   count good fits
    ;
    good_fit = 0
    ;
    for iline = 0, nlines-1 do begin
        ;
        ;  compute the nominal center channel for the line in the
        ;  input spectrum
        ;
        if(disp_in[2] ne 0.0) then begin
            col_cen = (-disp_in[1]/(2.0d0*disp_in[2])) + (1.0d0/(2.0d0*disp_in[2]))* $
                       sqrt(disp_in[1]^2 - (4.0d0*disp_in[2]*(disp_in[0]-abs(line_list[iline]))))
        endif else begin
           col_cen = 0.5d0 + (abs(line_list[iline]) - disp_in[0]) / disp_in[1]
        endelse
        ;
        line_col[iline] = col_cen
        ;if(idata eq 1) then print, "for line ",line_list[iline]," nominal center channel = ",col_cen
        ;
        ;  fit +/- fit_width channels about this center
        ;   previously =3
        ;   changed to 4 by gully on 20061023, better fitting
        ;
        ic_cen = ROUND(col_cen)
        ic1 = CEIL(col_cen + fit_width)
        ic2 = FLOOR(col_cen - fit_width)
        ;
        chmin = 0 > ic2
        chmax = (nspec-1) < ic1       ; fixed ncol-1 DPC 4/12/07
        nch = chmax - chmin + 1
        xgaus = float(indgen(nch) + chmin) + 0.5       ; center of channel
        ygaus = spec[chmin:chmax]
        ;
        ;   check for y all zeros
        ;
        ind_z = WHERE(ygaus ne 0 and ygaus ne -50.0, zcount)
        if(zcount eq 0) then begin
            col_cen = -100.
            a = [-1.0d6, -1.0e6, -1.0e6]
            sigma_a = [1.0e6, 1.0e6, 1.0e6]
            goto, loop_end
        endif
        ;
        ;   estimate center height and column
        ;
        if(line_abs[iline] eq 0) then max_s = max(ygaus) else max_s = min(ygaus)
        ;
        x_ind = WHERE(ygaus eq max_s)
        max_c = xgaus[x_ind[0]]
        width = 1.0
        offset = 0.0
        ;
        est = [max_s, max_c, width, offset]
        iter = 0
        ;
        loop:
        ;
        if(N_elements(ygaus) gt 4) then begin
        	yfit = wGAUSSFIT(xgaus,ygaus,a,NTERMS=4,SIGMA=sigma_a,ESTIMATES=est); , STATUS=status) - removed 20111130 DPC
        endif else begin
        	col_cen = -100.
            a = [-1.0d6, -1.0e6, -1.0e6]
            sigma_a = [1.0e6, 1.0e6, 1.0e6]
            goto, loop_end
        endelse
        ;
        ;   test for bad fits
        ;
        if(a[1] gt xgaus[nch-1] or a[1] lt xgaus[0]) then begin
            est[1] = max_c
            if(idata gt 0) then print,"..error found bad line center, resetting and refitting.. iter ",iter
            iter++
            if(iter lt 10) then goto, loop else begin
                    col_cen = -100.
                    a = [-1.0d6, -1.0e6, -1.0e6]
                   sigma_a = [1.0e6, 1.0e6, 1.0e6]
                    goto, loop_end
            endelse
        endif
        ;
        if(line_abs[iline] eq 0) then begin
	        if(a[0] lt 0. or a[2] lt 0.0) then begin
	            est[0] = max_s + 0.6*(iter-5)
	            est[2] = 1.0
	            if(idata gt 0) then print,"...error found bad line strength or width, resetting and refitting... iter ",iter
	            iter++
	            if(iter lt 10) then goto, loop else begin
	                col_cen = -100.
	                a = [-1.0d6, -1.0e6, -1.0e6]
	                sigma_a = [1.0e6, 1.0e6, 1.0e6]
	                goto, loop_end
	            endelse
	        endif
	    endif else begin
	    	if(a[0] gt 0. or a[2] lt 0.0) then begin
	            est[0] = max_s - 0.6*(iter-5)
	            est[2] = 1.0
	            if(idata gt 0) then print,"...error found bad line strength or width, resetting and refitting... iter ",iter
	            iter++
	            if(iter lt 10) then goto, loop else begin
	                col_cen = -100.
	                a = [-1.0d6, -1.0e6, -1.0e6]
	                sigma_a = [1.0e6, 1.0e6, 1.0e6]
	                goto, loop_end
	            endelse
	        endif
	    endelse
        ;
        ;loop_end:
        ;
        ;   write to the top window
        ;
        wset, window_numbers[0]
        ;erase
        ;
        ymin = min(ygaus)
        ymax = max(ygaus)
        dy = ymax-ymin
        ymin = ymin-0.1*dy
        ymax = ymax+0.2*dy
        dy = ymax - ymin
        ;
        plot, xgaus, ygaus, xrange=[chmin, chmax], yrange=[ymin, ymax], psym=4, xstyle=1, ystyle=1
        oplot, xgaus, yfit, color=12
        xyouts, col_cen+1, ymax-0.25*dy, STRING(format='(F6.4)',line_list[iline]), charsize=1.5
        ;
        ;  look for bad fits. These would have a[0] neg or < sigma_a[0],
        ;  or a[1] < 0 or > ncol, or widths < 0.1 cha or > 3 ch
        ;
        loop_end:
        ;
        if(line_abs[iline] eq 0) then begin
	        if(a[0] gt 0.0 and a[0] gt 0.5*sigma_a[0] and a[1] gt xgaus[0] and a[1] lt xgaus[nch-1] $
	            and a[2] gt 0.1 and a[2] lt fit_width) then begin
	           ;
	           ; save the fit
	           ;
	           fit_col[iline] = a[1]
	           fit_unc[iline] = sigma_a[1]
	           if(idata eq 1) then print, "Good fit for line ",line_list[iline]," at column ",a[1]," +/- ",sigma_a[1]
	           good_fit++
	        endif else begin
	           fit_col[iline] = 0.
	           fit_unc[iline] = 1000.
	           if(idata eq 1) then print, "Bad fit for line ",line_list[iline]," fit constants a[0-3] = ",a," and uncertainties = ",sigma_a
	        endelse
	    endif else begin
	    	if(a[0] lt 0.0 and a[0] lt 0.5*sigma_a[0] and a[1] gt xgaus[0] and a[1] lt xgaus[nch-1] $
	            and a[2] gt 0.1 and a[2] lt fit_width) then begin
	           ;
	           ; save the fit
	           ;
	           fit_col[iline] = a[1]
	           fit_unc[iline] = sigma_a[1]
	           if(idata eq 1) then print, "Good fit for line ",line_list[iline]," at column ",a[1]," +/- ",sigma_a[1]
	           good_fit++
	        endif else begin
	           fit_col[iline] = 0.
	           fit_unc[iline] = 1000.
	           if(idata eq 1) then print, "Bad fit for line ",line_list[iline]," fit constants a[0-3] = ",a," and uncertainties = ",sigma_a
	        endelse
	    endelse
        ;
        if(idata gt 1) then begin
            wait, 0.5
            cursor, xxx, yyy, /wait
        endif
    endfor
    ;
    ;   now that all lines have been fit, do regression if enough lines
    ;
    if(good_fit ge 5) then begin
        ;
        good_ind = WHERE(fit_col ne 0.0, good_count)
        if(good_count ge 4) then begin
           ;
           if(isave eq 1) then begin
               save_array[1,*] = fit_col[*]
               save_array[2,*] = fit_unc[*]
            endif
            ;
           line_list = abs(line_list[good_ind])
           fit_col = fit_col[good_ind]
           fit_unc = fit_unc[good_ind]
            ;
            ;   first, trap uncertainties to be no smaller than 0.1 pix
            ;
            fit_unc = 0.1 > fit_unc
            ;
            ;  load middle plot with dispersion fit of well fit lines
            ;
            wset, window_numbers[1]
            ;erase
            ;
            xmin = min(wave_limits)
            xmax = max(wave_limits)
            dx = xmax-xmin
            xmin = xmin-0.025*dx
            xmax = xmax+0.025*dx
            dx = xmax - xmin
            ;
            ymin = min(fit_col)
            ymax = max(fit_col)
            dy = ymax - ymin
            ymin = ymin - 0.15*dy
            ymax = ymax + 0.15*dy
            dy = ymax - ymin
            ;
            plot, line_list, fit_col, xrange=[xmin, xmax], yrange=[ymin, ymax], $
               xtitle="Calibration Line Wavelength [microns]",$
                 ytitle="Fitted Column Number", xstyle=1, ystyle=1, psym=4
            low = fit_col - 250.0*fit_unc
            high = fit_col + 250.0*fit_unc
            errplot, line_list, low, high, color=11
            ;
            xu = xmax-0.25*dx
            yu = ymin+0.25*dy
            ;
            xyouts, xu, yu, '250 X Unc.', color=11
            ;
            ;   perform regression of good lines
            ;
            wave_fit = POLY_FIT(line_list,fit_col,2,SIGMA=sigma, MEASURE_ERRORS = fit_unc, $
                   YFIT=yfit, CHISQ=chisq, /DOUBLE, STATUS=pstatus)
            resid = fit_col - yfit
            ;
            ;   model behavior
            ;
            xmodel = indgen(250)*(max(wave_limits)-min(wave_limits))/250.0 + min(wave_limits)
            ymodel = wave_fit[0] + wave_fit[1]*xmodel + wave_fit[2]*xmodel*xmodel
            ;
            oplot, xmodel, ymodel, color = 12
            ;
            ;   channel-based dispersion
            ;
            if(idata gt 0) then begin
                print," Column-Based Dispersion Equation: column = a[0] + a[1]*wave + a[2]*wave^2"
                for i=0,2 do begin
                    print," a[",i,"] = ",wave_fit[i]," +/- ",sigma[i]
                endfor
            endif
            ;
            if (idata gt 1) then begin
                wait, 0.5
                cursor, xxx, yyy, /WAIT
            endif
            ;
            ;   plot residuals to lowest plot
            ;
            wset, window_numbers[2]
            ;erase
            ;
            xlow = min(line_list)
            xhigh = max(line_list)
            xr = xhigh - xlow
            xlow = xlow - 0.1*xr
            xhigh = xhigh + 0.1*xr
            plot, line_list, resid, xrange=[xlow, xhigh], yrange=[-1.0, 1.5], xtitle="Calibration Line Wavelength [microns]",$
                 ytitle="Fit Residuals", xstyle=1, psym=4, ystyle=1
             ;
             ;  and errors
             ;
            erhigh = fit_unc + resid
            erlow = -1.0*fit_unc + resid
            errplot, line_list, erlow, erhigh, color=11
            ;
            xyouts,line_list, 0.75, STRING(format='(F6.4)',line_list), charsize=1.0, Orientation=90
            ;
            ;   zero line
            ;
            xx = [xlow, xhigh]
            yy = [0., 0.]
            oplot, xx, yy, color=12, linestyle=2
            ;
            ;   compute rms and add to plot
            ;
            ;       compute as the weighted rms deviations
            ;
            num = 0.0d0
            denom = 0.0d0
            for irms = 0, good_count-1 do begin
                num = num + (resid[irms]/fit_unc[irms])^2
                denom = denom + (1.0d0/fit_unc[irms])^2
            endfor
            ;
            rms = sqrt(num)/sqrt(denom)
            ;
            ;   scale so that more lines fit is better
            ;
            scale = float(nlines)/float(good_count)
            rms = rms * scale
            ;
            ;WCT_print_text, Event, STRING(format='("RMS_1 = ",f," pixels")',rms_1)
            ;
            ;
            ;zz = median_filtered_mean(resid)
            ;
            ;rms = STDDEV(resid)     ; don't use MFM, since it rejects outliers, which we shouldn't do here
            ;
            xyouts, xlow+0.5*xr, -0.5, STRING(format='("RMS = ",f6.2," pixels")',rms), charsize=1.5, alignment=0.0
            rms_wave = rms*disp_in[1]*10000.0
            xyouts, xlow+0.5*xr, -0.75, STRING(format='("      ",f9.4," Angstroms")',rms_wave), charsize=1.5, alignment=0.0
            ;
            ;   echo to GUI
            ;
            text_box_id = WIDGET_INFO( Event.top, FIND_BY_UNAME="S4_Refine_text_box")
            WIDGET_CONTROL, text_box_id, SET_VALUE=STRING(format='(f6.2)',rms)
            ;
            if(idata gt 1) then begin
                wait, 0.5
                cursor, xxx, yyy, /wait
            endif else wait, 0.06
            ;
            ;   calc mean resid
            ;
            zz_resid = median_filtered_mean(resid)
            ;
            ;   compute unc for wavelengths
            ;
            s_wave1 = fltarr(good_count)
            s_wave1[*] = zz_resid[1]/wave_fit[1]
            ;
            ;   use slope term times column uncertainty from line fits
            ;
            s_wave2 = disp_in[1] * fit_unc
            ;
            s_wave = fltarr(good_count)
            for kk = 0, good_count-1 do begin
                s_wave[kk] = s_wave1[kk] > s_wave2[kk]
            endfor
            ;
            if(idata eq 1) then begin
                print, "resid rms = ",zz_resid[1]," column pixels"
                print, "s_wave = ",s_wave
            endif
            ;
            ;   wavelength-based dispersion equation
            ;
            disp_out = POLY_FIT(fit_col,line_list,2,SIGMA=sigma, MEASURE_ERRORS = s_wave, YFIT=yfit, $
                   CHISQ=chisq, /DOUBLE, STATUS=dstatus)
            ;
       endif else begin
              rms = 1000.0
            disp_out = disp_in
       endelse
    endif else begin
       ;
       ; if too few good line fits were obtained, set the RMS to high value, load disp_out with disp_in
       ;     and return
       ;
       rms = 1000.0
       disp_out = disp_in
    endelse
    ;
    ;
    if(idata eq 1 and good_fit ge 5) then begin
        print, "Wavelength-based Dispersion Equation: wave = b[0] + b[1] * col + b[2] * col^2"
        for i=0,2 do begin
            print," b[",i,"] = ",disp_out[i]," +/- ",sigma[i]
        endfor
    endif
    ;
    ;   replot, with new dispersion information
    ;
    ;   plot the spectrum from the start, stop limits and the locations of all the
    ;   cal lines into the base plot region
    ;
    trial_wave = disp_out[0] + disp_out[1]*x_spec + disp_out[2]*x_spec*x_spec
    ind_trial = WHERE(trial_wave ge min(wave_limits) and trial_wave le max(wave_limits),ntrial)
    if(ntrial gt 0 and noplot eq 0) then begin
        trial_wave = trial_wave[ind_trial]
        trial_spec = spec[ind_trial]
        ;
        WCT_plot_spectrum_and_cal_lines, Event, trial_wave, trial_spec, lines, line_qual, $
            cal_mode
        ;
    endif
    ;
    if(idata gt 0) then begin
            print, "Click cursor to continue"
            wait, 0.5
            cursor, xxx, yyy, /wait
    endif
    ;
endif else begin
    WCT_print_text, Event, "No Calibration Lines within Limits. Fit Not Performed."
endelse
;
end

